#include "mruby.h"
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include "error.h"
#include "opcode.h"
#include "irep.h"
#include "mproc.h"
#include "mnumeric.h"
#include "variable.h"
#include "mstring.h"
#include "eval_intern.h"
#include "mclass.h"

#define warn_printf printf

#ifndef FALSE
#define FALSE   0
#endif

#ifndef TRUE
#define TRUE    1
#endif

mrb_value
mrb_exc_new(mrb_state *mrb, struct RClass *c, const char *ptr, long len)
{
  return mrb_funcall(mrb, mrb_obj_value(c), "new", 1, mrb_str_new(mrb, ptr, len));
}

mrb_value
mrb_exc_new3(mrb_state *mrb, struct RClass* c, mrb_value str)
{
  //StringValue(str);
  mrb_string_value(mrb, &str);
  return mrb_funcall(mrb, mrb_obj_value(c), "new", 1, str);
}

//mrb_value make_exception(mrb_state *mrb, int argc, mrb_value *argv, int isstr);

void
mrb_raise(mrb_state *mrb, struct RClass *c, const char *fmt, ...)
{
  va_list args;
  char buf[256];
  mrb_code* pcode;
  int i;

  va_start(args, fmt);
  vsnprintf(buf, 256, fmt, args);
  va_end(args);
  mrb->exc = (struct RObject*)c;
  struct RRescue *r = mrb_obj_alloc(mrb, MRB_TT_RESCUE, 0);
  pcode = mrb->ci->pc;
  for (i=0;i<mrb->ci->proc->body.irep->ilen;i++){
    if (GET_OPCODE(mrb->ci->proc->body.irep->iseq[i]) == OP_STOP) {
      pcode = &mrb->ci->proc->body.irep->iseq[i];
      break;
    }
  }
  r->next = mrb->rq;
  r->pc = pcode;
  mrb->rq = r;
  mrb->ci->rq = mrb->rq;
  printf("%s\n", buf);
  exit(1);
}

void
rb_raise(struct RClass *c, const char *fmt, ...)
{
  va_list args;
  char buf[256];

  va_start(args, fmt);
  vsnprintf(buf, 256, fmt, args);
  va_end(args);
  //mrb_exc_raise(mrb_exc_new2(exc, buf));
  printf("%s", buf);
}

void
mrb_name_error(mrb_state *mrb, mrb_sym id, const char *fmt, ...)
{
  mrb_value exc, argv[2];
  va_list args;
  char buf[256];

  va_start(args, fmt);
  //argv[0] = mrb_vsprintf(fmt, args);
  vsnprintf(buf, 256, fmt, args);
  argv[0] = mrb_str_new(mrb, buf, strlen(buf));
  va_end(args);

  argv[1] = mrb_str_new_cstr(mrb, mrb_sym2name(mrb, id));
  exc = mrb_class_new_instance(mrb, 2, argv, mrb->eNameError_class);
  mrb_exc_raise(mrb, exc);
}
mrb_value
mrb_sprintf(mrb_state *mrb, const char *fmt, ...)
{
  va_list args;
  char buf[256];

  va_start(args, fmt);
  vsnprintf(buf, 256, fmt, args);
  va_end(args);
  return mrb_str_new(mrb, buf, strlen(buf));
}

void
mrb_warn(const char *fmt, ...)
{
  va_list args;
  char buf[256];

  va_start(args, fmt);
  snprintf(buf, 256, "warning: %s", fmt);
  printf(buf, args);
  va_end(args);
}


void
mrb_warning(const char *fmt, ...)
{
  va_list args;
  char buf[256];

  va_start(args, fmt);
  snprintf(buf, 256, "warning: %s", fmt);
  printf(buf, args);
  va_end(args);
}

void
mrb_bug(const char *fmt, ...)
{
  va_list args;
  char buf[256];

  va_start(args, fmt);
  snprintf(buf, 256, "bug: %s", fmt);
  printf(buf, args);
  va_end(args);
}

static const char *
mrb_strerrno(int err)
{
#define defined_error(name, num) if (err == num) return name;
#define undefined_error(name)
//#include "known_errors.inc"
#undef defined_error
#undef undefined_error
    return NULL;
}

void
mrb_bug_errno(const char *mesg, int errno_arg)
{
  if (errno_arg == 0)
    mrb_bug("%s: errno == 0 (NOERROR)", mesg);
  else {
    const char *errno_str = mrb_strerrno(errno_arg);
    if (errno_str)
      mrb_bug("%s: %s (%s)", mesg, strerror(errno_arg), errno_str);
    else
      mrb_bug("%s: %s (%d)", mesg, strerror(errno_arg), errno_arg);
  }
}

void error_print(void)
{
  printf("error_print\n");
}

int
sysexit_status(mrb_state *mrb, mrb_value err)
{
  mrb_value st = mrb_iv_get(mrb, err, mrb_intern(mrb, "status"));
  return mrb_fixnum(st);
}

void
error_pos(void)
{
#if 0
  const char *sourcefile = mrb_sourcefile();
  int sourceline = mrb_sourceline();

  if (sourcefile) {
    if (sourceline == 0) {
      warn_printf("%s", sourcefile);
    }
    else if (mrb_frame_callee()) {
      warn_printf("%s:%d:in `%s'", sourcefile, sourceline,
                mrb_sym2name(mrb, mrb_frame_callee()));
    }
    else {
      warn_printf("%s:%d", sourcefile, sourceline);
    }
  }
#endif
}

static void
set_backtrace(mrb_state *mrb, mrb_value info, mrb_value bt)
{
	mrb_funcall(mrb, info, "set_backtrace", 1, bt);
}

mrb_value
make_exception(mrb_state *mrb, int argc, mrb_value *argv, int isstr)
{
  mrb_value mesg;
  int n;

  mesg = mrb_nil_value();
  switch (argc) {
    case 0:
    break;
    case 1:
      if (mrb_nil_p(argv[0]))
        break;
      if (isstr) {
        mesg = mrb_check_string_type(mrb, argv[0]);
        if (!mrb_nil_p(mesg)) {
          mesg = mrb_exc_new3(mrb, mrb->eRuntimeError_class, mesg);
          break;
        }
      }
      n = 0;
      goto exception_call;

    case 2:
    case 3:
      n = 1;
exception_call:
      //if (argv[0] == sysstack_error) return argv[0];

      //CONST_ID(mrb, exception, "exception");
      //mesg = mrb_check_funcall(mrb, argv[0], exception, n, argv+1);
      //if (mrb_nil_p(mesg)) {
      //  /* undef */
      //  mrb_raise(mrb, mrb->eTypeError_class, "exception class/object expected");
      //}
      if (mrb_respond_to(mrb, argv[0], mrb_intern(mrb, "exception"))) {
        mrb_funcall(mrb, argv[0], "exception", n, argv+1);
      }
      else {
        /* undef */
        mrb_raise(mrb, mrb->eTypeError_class, "exception class/object expected");
      }

      break;
    default:
      mrb_raise(mrb, mrb->eArgumentError_class, "wrong number of arguments (%d for 0..3)", argc);
      break;
  }
  if (argc > 0) {
    if (!mrb_obj_is_kind_of(mrb, mesg, MRB_TT_EXCEPTION, mrb->eException_class))
      mrb_raise(mrb, mrb->eTypeError_class, "exception object expected");
    if (argc > 2)
        set_backtrace(mrb, mesg, argv[2]);
  }

  return mesg;
}

mrb_value
mrb_make_exception(mrb_state *mrb, int argc, mrb_value *argv)
{
  return make_exception(mrb, argc, argv, TRUE);
}

void
mrb_sys_fail(const char *mesg)
{
  //mrb_exc_raise(make_errno_exc(mesg));
  printf("%s", mesg);
}

void
mrb_init_exception(mrb_state *mrb)
{
  mrb->eException_class         = mrb_define_class(mrb, "Exception",           mrb->object_class);         /* 15.2.22 */

  mrb->eStandardError_class     = mrb_define_class(mrb, "StandardError",       mrb->eException_class);     /* 15.2.23 */

#ifndef SYMPLE_ERROR_CLASS // added by yamashita

  mrb->eTypeError_class         = mrb_define_class(mrb, "TypeError",           mrb->eStandardError_class); /* 15.2.29 */
  mrb->eArgumentError_class     = mrb_define_class(mrb, "ArgumentError",       mrb->eStandardError_class); /* 15.2.24 */
  mrb->eIndexError_class        = mrb_define_class(mrb, "IndexError",          mrb->eStandardError_class); /* 15.2.33 */
  mrb->eRangeError_class        = mrb_define_class(mrb, "RangeError",          mrb->eStandardError_class); /* 15.2.26 */
  mrb->eNameError_class         = mrb_define_class(mrb, "NameError",           mrb->eStandardError_class); /* 15.2.31 */

  mrb->eNoMethodError_class     = mrb_define_class(mrb, "NoMethodError",       mrb->eNameError_class);     /* 15.2.32 */
  mrb->eScriptError_class       = mrb_define_class(mrb, "ScriptError",         mrb->eException_class);     /* 15.2.37 */
  mrb->eSyntaxError_class       = mrb_define_class(mrb, "SyntaxError",         mrb->eScriptError_class);   /* 15.2.38 */
  mrb->eLoadError_class         = mrb_define_class(mrb, "LoadError",           mrb->eScriptError_class);   /* 15.2.39 */
  mrb->eRuntimeError_class      = mrb_define_class(mrb, "RuntimeError",        mrb->eStandardError_class); /* 15.2.28 */
  mrb->eSystemCallError_class   = mrb_define_class(mrb, "SystemCallError",     mrb->eStandardError_class); /* 15.2.36 */
  mrb->eLocalJumpError_class    = mrb_define_class(mrb, "LocalJumpError",      mrb->eStandardError_class); /* 15.2.25 */
  mrb->eRegexpError_class       = mrb_define_class(mrb, "RegexpError",         mrb->eStandardError_class); /* 15.2.27 */
  mrb->eZeroDivisionError_class = mrb_define_class(mrb, "ZeroDivisionError",   mrb->eStandardError_class); /* 15.2.30 */

  mrb->eIOError_class           = mrb_define_class(mrb, "IOError",             mrb->eStandardError_class); /* 15.2.34 */
  mrb->eEOFError_class          = mrb_define_class(mrb, "EOFError",            mrb->eIOError_class);       /* 15.2.35 */
  mrb->eEncodingError_class     = mrb_define_class(mrb, "EncodingError", mrb->eStandardError_class);
  mrb->eNotImpError_class       = mrb_define_class(mrb, "NotImplementedError", mrb->eScriptError_class);

  mrb->eSystemExit_class        = mrb_define_class(mrb, "SystemExit",          mrb->eException_class);
#endif
}
