/************************************************

  fcntl.c -

  $Author: nobu $
  created at: Mon Apr  7 18:53:05 JST 1997

  Copyright (C) 1997-2001 Yukihiro Matsumoto

************************************************/

/************************************************
= NAME

fcntl - load the C fcntl.h defines

= SYNOPSIS

    require "fcntl"
    m = s.fcntl(Fcntl::F_GETFL, 0)
    f.fcntl(Fcntl::F_SETFL, Fcntl::O_NONBLOCK|m)

= DESCRIPTION

This module is just a translation of the C <fcntl.h> file.

= NOTE

Only #define symbols get translated; you must still correctly
pack up your own arguments to pass as args for locking functions, etc.

************************************************/

#include "mruby.h"
#include <fcntl.h>
#include "mio.h"

/* Fcntl loads the constants defined in the system's <fcntl.h> C header
 * file, and used with both the fcntl(2) and open(2) POSIX system calls.
 *
 * Copyright (C) 1997-2001 Yukihiro Matsumoto
 *
 * Documented by mathew <meta@pobox.com>
 *
 * = Usage
 *
 * To perform a fcntl(2) operation, use IO::fcntl in the core classes.
 *
 * To perform an open(2) operation, use IO::sysopen.
 *
 * The set of operations and constants available depends upon specific OS
 * platform. Some values listed below may not be supported on your system.
 *
 * The constants supported by Ruby for use with IO::fcntl are:
 *
 * - F_DUPFD - duplicate a close-on-exec file handle to a non-close-on-exec
 *   file handle.
 *
 * - F_GETFD - read the close-on-exec flag of a file handle.
 *
 * - F_SETFD - set the close-on-exec flag of a file handle.
 *
 * - FD_CLOEXEC - the value of the close-on-exec flag.
 *
 * - F_GETFL - get file descriptor flags.
 *
 * - F_SETFL - set file descriptor flags.
 *
 * - O_APPEND, O_NONBLOCK, etc (see below) - file descriptor flag
 *   values for the above.
 *
 * - F_GETLK - determine whether a given region of a file is locked.
 *
 * - F_SETLK - acquire a lock on a region of a file.
 *
 * - F_SETLKW - acquire a lock on a region of a file, waiting if necessary.
 *
 * - F_RDLCK, F_WRLCK, F_UNLCK - types of lock for the above.
 *
 * The constants supported by Ruby for use with IO::sysopen are:
 *
 * - O_APPEND - open file in append mode.
 *
 * - O_NOCTTY - open tty without it becoming controlling tty.
 *
 * - O_CREAT - create file if it doesn't exist.
 *
 * - O_EXCL - used with O_CREAT, fail if file exists.
 *
 * - O_TRUNC - truncate file on open.
 *
 * - O_NONBLOCK / O_NDELAY - open in non-blocking mode.
 *
 * - O_RDONLY - open read-only.
 *
 * - O_WRONLY - open write-only.
 *
 * - O_RDWR - open read-write.
 *
 * - O_ACCMODE - mask to extract read/write flags.
 *
 * Example:
 *
 *   require 'fcntl'
 *
 *   fd = IO::sysopen('/tmp/tempfile',
 *        Fcntl::O_WRONLY | Fcntl::O_EXCL | Fcntl::O_CREAT)
 *   f = IO.open(fd)
 *   f.syswrite("TEMP DATA")
 *   f.close
 *
 */
void
mrb_init_fcntl(mrb_state *mrb)
{
  struct RClass *mFcntl;

  mFcntl = mrb_define_module(mrb, "Fcntl");
#ifdef F_DUPFD
  mrb_define_const(mrb, mFcntl, "F_DUPFD", mrb_fixnum_value(F_DUPFD));
#endif
#ifdef F_GETFD
  mrb_define_const(mrb, mFcntl, "F_GETFD", mrb_fixnum_value(F_GETFD));
#endif
#ifdef F_GETLK
  mrb_define_const(mrb, mFcntl, "F_GETLK", mrb_fixnum_value(F_GETLK));
#endif
#ifdef F_SETFD
  mrb_define_const(mrb, mFcntl, "F_SETFD", mrb_fixnum_value(F_SETFD));
#endif
#ifdef F_GETFL
  mrb_define_const(mrb, mFcntl, "F_GETFL", mrb_fixnum_value(F_GETFL));
#endif
#ifdef F_SETFL
  mrb_define_const(mrb, mFcntl, "F_SETFL", mrb_fixnum_value(F_SETFL));
#endif
#ifdef F_SETLK
  mrb_define_const(mrb, mFcntl, "F_SETLK", mrb_fixnum_value(F_SETLK));
#endif
#ifdef F_SETLKW
  mrb_define_const(mrb, mFcntl, "F_SETLKW", mrb_fixnum_value(F_SETLKW));
#endif
#ifdef FD_CLOEXEC
  mrb_define_const(mrb, mFcntl, "FD_CLOEXEC", mrb_fixnum_value(FD_CLOEXEC));
#endif
#ifdef F_RDLCK
  mrb_define_const(mrb, mFcntl, "F_RDLCK", mrb_fixnum_value(F_RDLCK));
#endif
#ifdef F_UNLCK
  mrb_define_const(mrb, mFcntl, "F_UNLCK", mrb_fixnum_value(F_UNLCK));
#endif
#ifdef F_WRLCK
  mrb_define_const(mrb, mFcntl, "F_WRLCK", mrb_fixnum_value(F_WRLCK));
#endif
#ifdef O_CREAT
  mrb_define_const(mrb, mFcntl, "O_CREAT", mrb_fixnum_value(O_CREAT));
#endif
#ifdef O_EXCL
  mrb_define_const(mrb, mFcntl, "O_EXCL", mrb_fixnum_value(O_EXCL));
#endif
#ifdef O_NOCTTY
  mrb_define_const(mrb, mFcntl, "O_NOCTTY", mrb_fixnum_value(O_NOCTTY));
#endif
#ifdef O_TRUNC
  mrb_define_const(mrb, mFcntl, "O_TRUNC", mrb_fixnum_value(O_TRUNC));
#endif
#ifdef O_APPEND
  mrb_define_const(mrb, mFcntl, "O_APPEND", mrb_fixnum_value(O_APPEND));
#endif
#ifdef O_NONBLOCK
  mrb_define_const(mrb, mFcntl, "O_NONBLOCK", mrb_fixnum_value(O_NONBLOCK));
#endif
#ifdef O_NDELAY
  mrb_define_const(mrb, mFcntl, "O_NDELAY", mrb_fixnum_value(O_NDELAY));
#endif
#ifdef O_RDONLY
  mrb_define_const(mrb, mFcntl, "O_RDONLY", mrb_fixnum_value(O_RDONLY));
#endif
#ifdef O_RDWR
  mrb_define_const(mrb, mFcntl, "O_RDWR", mrb_fixnum_value(O_RDWR));
#endif
#ifdef O_WRONLY
  mrb_define_const(mrb, mFcntl, "O_WRONLY", mrb_fixnum_value(O_WRONLY));
#endif
#ifdef O_ACCMODE
  mrb_define_const(mrb, mFcntl, "O_ACCMODE", mrb_fixnum_value(O_ACCMODE));
#else
  mrb_define_const(mrb, mFcntl, "O_ACCMODE", mrb_fixnum_value(O_RDONLY | O_WRONLY | O_RDWR));
#endif
}
