#include "mruby.h"
#include <stdlib.h>
#include <string.h>
#include "class.h"
#include "range.h"
#include "variable.h"
#include "error.h"
#include "numeric.h"
#include "mio.h"
#include "mstring.h"
#include "hash.h"

#ifdef _WIN32
#include "missing/file.h"
#endif
#ifdef __CYGWIN__
#include <windows.h>
#include <sys/cygwin.h>
#endif

#ifdef HAVE_SYS_FILE_H
# include <sys/file.h>
#else
int flock(int, int);
#endif

#ifndef MAXPATHLEN
# define MAXPATHLEN 1024
#endif

#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <time.h>

/* --------------------------------------------- */
/* define system APIs */
//#ifdef _WIN32
#if 0
#define STAT(p, s)        rb_w32_ustati64(p, s)
#undef lstat
#define lstat(p, s)        rb_w32_ustati64(p, s)
#undef access
#define access(p, m)        rb_w32_uaccess(p, m)
#undef chmod
#define chmod(p, m)        rb_w32_uchmod(p, m)
#undef chown
#define chown(p, o, g)        rb_w32_uchown(p, o, g)
#undef utime
#define utime(p, t)        rb_w32_uutime(p, t)
#undef link
#define link(f, t)        rb_w32_ulink(f, t)
#undef unlink
#define unlink(p)        rb_w32_uunlink(p)
#undef rename
#define rename(f, t)        rb_w32_urename(f, t)
#else
#define STAT(p, s)        stat(p, s)
#endif

//#define insecure_obj_p(obj, level) (level >= 4 || (level > 0 && OBJ_TAINTED(obj)))
#define insecure_obj_p(obj, level) (0)

#ifndef FALSE
#define FALSE   0
#endif

#ifndef TRUE
#define TRUE    1
#endif

#ifdef __CYGWIN__
#undef DOSISH
#endif
#if defined(DOSISH)
#define PATH_SEP ";"
#else
#define PATH_SEP ":"
#endif
#define PATH_SEP_CHAR PATH_SEP[0]
#define PATH_ENV "PATH"
#define my_getcwd() ruby_getcwd()
#define HAVE_GETCWD 1
/* --------------------------------------------- */
static mrb_value
file_path_convert(mrb_state *mrb, mrb_value name)
{
#ifdef INCLUDE_REGEXP
  #ifndef _WIN32 /* non Windows == Unix */
    rb_encoding *fname_encoding = rb_enc_from_index(mrb, ENCODING_GET(mrb, name));
    rb_encoding *fs_encoding;
    if (rb_default_internal_encoding(mrb) != NULL
       && rb_usascii_encoding(mrb) != fname_encoding
       && rb_ascii8bit_encoding(mrb) != fname_encoding
       && (fs_encoding = rb_filesystem_encoding(mrb)) != fname_encoding) {
       /* Don't call rb_filesystem_encoding() before US-ASCII and ASCII-8BIT */
       name = rb_str_conv_enc(mrb, name, fname_encoding, fs_encoding);
    }
  #endif
#endif
    return name;
}

static mrb_value
rb_get_path_check(mrb_state *mrb, mrb_value obj, int level)
{
  mrb_value tmp;
  //mrb_sym to_path;

  //if (insecure_obj_p(obj, level)) {
  //  rb_insecure_operation();
  //}

  //CONST_ID(mrb, to_path, "to_path");
  //tmp = rb_check_funcall(mrb, obj, to_path, 0, 0);
  //if (tmp == Qundef) {
  //  tmp = obj;
  //}
  if (mrb_respond_to(mrb, obj, mrb_intern(mrb, "to_path"))) {
    mrb_funcall(mrb, obj, "to_path", 0);
  }
  else {
    tmp = obj;
  }
  //StringValue(tmp);
  mrb_string_value(mrb, &tmp);

  tmp = file_path_convert(mrb, tmp);
  //StringValueCStr(tmp);
  rb_string_value_cstr(mrb, &tmp);
  //if ((!mrb_obj_equal(mrb, obj, tmp)) && insecure_obj_p(tmp, level)) {
  //  rb_insecure_operation();
  //}
#ifdef INCLUDE_REGEXP
  rb_enc_check(mrb, tmp, rb_enc_from_encoding(mrb, rb_usascii_encoding(mrb)));
#endif
  return rb_str_new4(mrb, tmp);
}

mrb_value
rb_get_path(mrb_state *mrb, mrb_value obj)
{
  return rb_get_path_check(mrb, obj, 0/*rb_safe_level()*/);
}

#ifdef INCLUDE_REGEXP
mrb_value
rb_str_encode_ospath(mrb_state *mrb, mrb_value path)
{
 #ifdef _WIN32
  rb_encoding *enc = rb_enc_get(mrb, path);
  if (enc != rb_ascii8bit_encoding(mrb)) {
    rb_encoding *utf8 = rb_utf8_encoding(mrb);
    if (enc != utf8)
        path = rb_str_encode(mrb, path, rb_enc_from_encoding(mrb, utf8), 0, mrb_nil_value());
  }
  else if (RSTRING_LEN(path) > 0) {
    path = mrb_str_dup(mrb, path);
    rb_enc_associate(mrb, path, rb_filesystem_encoding(mrb));
    path = rb_str_encode(mrb, path, rb_enc_from_encoding(mrb, rb_utf8_encoding(mrb)), 0, mrb_nil_value());
  }
 #endif
  return path;
}
#endif

static int
mrb_stat(mrb_state *mrb, mrb_value file, struct stat *st)
{
  mrb_value tmp;

  //rb_secure(2);
  tmp = mrb_check_convert_type(mrb, file, MRB_TT_FILE, "IO", "to_io");
  if (!mrb_nil_p(tmp)) {
    rb_io_t *fptr;

    GetOpenFile(mrb, tmp, fptr);
#ifdef INCLUDE_REGEXP
    return fstat(fptr->fd, st);
#else
    return fstat(fileno(fptr->f), st);
#endif
  }
#ifdef INCLUDE_REGEXP
  FilePathValue(mrb, file);
  file = rb_str_encode_ospath(mrb, file);
#else
  //SafeStringValue(file);
  mrb_string_value(mrb, &file);
#endif
  //return STAT(StringValueCStr(file), st);
  char* tmpStr = rb_string_value_cstr(mrb, &file);
  return STAT(tmpStr, st);
}

#define strrdirsep rb_path_last_separator
#if defined __CYGWIN__ || defined DOSISH
#define DOSISH_UNC
#define DOSISH_DRIVE_LETTER
#define FILE_ALT_SEPARATOR '\\'
#endif
#ifdef FILE_ALT_SEPARATOR
#define isdirsep(x) ((x) == '/' || (x) == FILE_ALT_SEPARATOR)
static const char file_alt_separator[] = {FILE_ALT_SEPARATOR, '\0'};
#else
#define isdirsep(x) ((x) == '/')
#endif

#ifndef USE_NTFS
#if defined _WIN32 || defined __CYGWIN__
#define USE_NTFS 1
#else
#define USE_NTFS 0
#endif
#endif

#if USE_NTFS
#define istrailinggarbage(x) ((x) == '.' || (x) == ' ')
#else
#define istrailinggarbage(x) 0
#endif

#ifndef CharNext		/* defined as CharNext[AW] on Windows. */
# define CharNext(p) ((p) + 1)
#endif

char *
rb_path_last_separator(const char *path)
{
    char *last = NULL;
    while (*path) {
        if (isdirsep(*path)) {
            const char *tmp = path++;
            while (isdirsep(*path)) path++;
            if (!*path) break;
            last = (char *)tmp;
        }
        else {
            path = CharNext(path);
        }
    }
    return last;
}

char *
ruby_getcwd(void)
{
#ifdef HAVE_GETCWD
    int size = 200;
    char *buf = xmalloc(size);

    while (!getcwd(buf, size)) {
        if (errno != ERANGE) {
            xfree(buf);
            rb_sys_fail("getcwd");
        }
        size *= 2;
        buf = xrealloc(buf, size);
    }
#else
# ifndef PATH_MAX
#  define PATH_MAX 8192
# endif
    char *buf = xmalloc(PATH_MAX+1);

    if (!getwd(buf)) {
        xfree(buf);
        rb_sys_fail("getwd");
    }
#endif
    return buf;
}

int
rb_is_absolute_path(const char *path)
{
#ifdef DOSISH_DRIVE_LETTER
    if (has_drive_letter(path) && isdirsep(path[2])) return 1;
#endif
#ifdef DOSISH_UNC
    if (isdirsep(path[0]) && isdirsep(path[1])) return 1;
#endif
#ifndef DOSISH
    if (path[0] == '/') return 1;
#endif
    return 0;
}

#ifndef ENABLE_PATH_CHECK
# if defined DOSISH || defined __CYGWIN__
#   define ENABLE_PATH_CHECK 0
# else
#   define ENABLE_PATH_CHECK 1
# endif
#endif

#if ENABLE_PATH_CHECK
static int
path_check_0(mrb_state *mrb, mrb_value path, int execpath)
{
    struct stat st;
    //const char *p0 = StringValueCStr(path);
    const char *p0 = rb_string_value_cstr(mrb, &path);
    char *p = 0, *s;

    if (!rb_is_absolute_path(p0)) {
        char *buf = my_getcwd();
        mrb_value newpath;

        newpath = mrb_str_new_cstr(mrb, buf);//rb_str_new2(buf);
        xfree(buf);

        rb_str_cat2(mrb, newpath, "/");
        rb_str_cat2(mrb, newpath, p0);
        path = newpath;
        p0 = RSTRING_PTR(path);
    }
    for (;;) {
#ifndef S_IWOTH
# define S_IWOTH 002
#endif
        if (STAT(p0, &st) == 0 && S_ISDIR(st.st_mode) && (st.st_mode & S_IWOTH)
#ifdef S_ISVTX
            && !(p && execpath && (st.st_mode & S_ISVTX))
#endif
            && !access(p0, W_OK)) {
            rb_warn("Insecure world writable dir %s in %sPATH, mode 0%o",
                    p0, (execpath ? "" : "LOAD_"), st.st_mode);
            if (p) *p = '/';
            //RB_GC_GUARD(path);
            return 0;
        }
        s = strrdirsep(p0);
        if (p) *p = '/';
        if (!s || s == p0) return 1;
        p = s;
        *p = '\0';
    }
}
#endif

#if ENABLE_PATH_CHECK
#define fpath_check(path) path_check_0(path, FALSE)
#else
#define fpath_check(path) 1
#endif

int
rb_path_check(mrb_state *mrb, const char *path)
{
#if ENABLE_PATH_CHECK
    const char *p0, *p, *pend;
    const char sep = PATH_SEP_CHAR;

    if (!path) return 1;

    pend = path + strlen(path);
    p0 = path;
    p = strchr(path, sep);
    if (!p) p = pend;

    for (;;) {
        if (!path_check_0(mrb, mrb_str_new(mrb, p0, p - p0), TRUE)) {
            return 0;                /* not safe */
        }
        p0 = p + 1;
        if (p0 > pend) break;
        p = strchr(p0, sep);
        if (!p) p = pend;
    }
#endif
    return 1;
}

/* 15.2.21.3.1  */
/*
 * call-seq:
 *    File.exist?(file_name)    ->  true or false
 *    File.exists?(file_name)   ->  true or false
 *
 * Return <code>true</code> if the named file exists.
 */

static mrb_value
mrb_file_exist_p(mrb_state *mrb, mrb_value obj)
{
  mrb_value fname;
  struct stat st;

  mrb_get_args(mrb, "o", &fname);

  if (mrb_stat(mrb, fname, &st) < 0) return mrb_false_value();
  return mrb_true_value();
}

/* 15.2.21.4.1  */
/*
 *  call-seq:
 *     File.new(filename, mode="r" [, opt])            -> file
 *     File.new(filename [, mode [, perm]] [, opt])    -> file
 *
 *  Opens the file named by _filename_ according to
 *  _mode_ (default is ``r'') and returns a new
 *  <code>File</code> object.
 *
 *  === Parameters
 *  See the description of class +IO+ for a description of _mode_.
 *  The file mode may optionally be specified as a +Fixnum+
 *  by _or_-ing together the flags (O_RDONLY etc,
 *  again described under +IO+).
 *
 *  Optional permission bits may be given in _perm_.
 *  These mode and permission bits are platform dependent;
 *  on Unix systems, see <code>open(2)</code> for details.
 *
 *  Optional _opt_ parameter is same as in <code.IO.open</code>.
 *
 *  === Examples
 *
 *     f = File.new("testfile", "r")
 *     f = File.new("newfile",  "w+")
 *     f = File.new("newfile", File::CREAT|File::TRUNC|File::RDWR, 0644)
 */

static mrb_value
mrb_file_initialize(mrb_state *mrb, mrb_value self)
{
  mrb_value argv[16];
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
  if (argc < 3) argv[2] = mrb_nil_value();
  if (argc < 2) argv[1] = mrb_nil_value();
  if (RFILE(self)->fptr) {
    mrb_raise(mrb, mrb->eRuntimeError_class, "reinitializing File");
  }
  if (argc == 1 || argc == 2) {
    mrb_value fd = mrb_check_convert_type(mrb, argv[0], MRB_TT_FIXNUM, "Fixnum", "to_int");

    if (!mrb_nil_p(fd)) {
      /* first argument is file discriptor */
      argv[0] = fd;
      return rb_io_initialize(mrb, argc, argv, self);
    }
  }
  rb_open_file(mrb, argc, argv, self);

  return self;
}

/* 15.2.21.4.2  */
/*
 *  call-seq:
 *     file.path  ->  filename
 *
 *  Returns the pathname used to create <i>file</i> as a string. Does
 *  not normalize the name.
 *
 *     File.new("testfile").path               #=> "testfile"
 *     File.new("/tmp/../tmp/xxx", "w").path   #=> "/tmp/../tmp/xxx"
 *
 */

static mrb_value
mrb_file_path(mrb_state *mrb, mrb_value self)
{
    rb_io_t *fptr;

    fptr = RFILE(self)->fptr;
    mrb_io_check_initialized(mrb, fptr);
#ifdef INCLUDE_REGEXP
    if (mrb_nil_p(fptr->pathv)) return mrb_nil_value();
    //return rb_obj_taint(rb_str_dup(fptr->pathv));
    return mrb_str_dup(mrb, fptr->pathv);
#else
    if (!fptr->path) return mrb_nil_value();
    return mrb_str_new_cstr(mrb, fptr->path);
#endif
}

void
mrb_init_file(mrb_state *mrb)
{
  struct RClass *c;
  c = mrb->file_class = mrb_define_class(mrb, "File", mrb->io_class); /* class File < IO */
  mrb_define_class_method(mrb, c, "exist?",     mrb_file_exist_p,    ARGS_REQ(1));              /* 15.2.21.3.1  */
  mrb_define_method(mrb, c,       "initialize", mrb_file_initialize, ARGS_ANY());               /* 15.2.21.4.1  */
  mrb_define_method(mrb, c,       "path",       mrb_file_path,       ARGS_NONE());              /* 15.2.21.4.2  */
}
