/************************************************

  basicsocket.c -

  created at: Thu Mar 31 12:21:29 JST 1994

  Copyright (C) 1993-2007 Yukihiro Matsumoto

************************************************/

#include "rubysocket.h"

static VALUE
bsock_close(mrb_state *mrb, mrb_value self)
{
  mrb_value fd;

  fd = mrb_iv_get(mrb, self, mrb_intern(mrb, "socket_fd"));
  if (close(mrb_fixnum(fd)) == -1)
    mrb_sys_fail("close");

  return mrb_fixnum_value(0);
}

/*
 * call-seq:
 *   basicsocket.shutdown([how]) => 0
 *
 * Calls shutdown(2) system call.
 *
 * s.shutdown(Socket::SHUT_RD) disallows further read.
 *
 * s.shutdown(Socket::SHUT_WR) disallows further write.
 *
 * s.shutdown(Socket::SHUT_RDWR) disallows further read and write.
 *
 * _how_ can be symbol or string:
 * - :RD, :SHUT_RD, "RD" and "SHUT_RD" are accepted as Socket::SHUT_RD.
 * - :WR, :SHUT_WR, "WR" and "SHUT_WR" are accepted as Socket::SHUT_WR.
 * - :RDWR, :SHUT_RDWR, "RDWR" and "SHUT_RDWR" are accepted as Socket::SHUT_RDWR.
 *
 *   UNIXSocket.pair {|s1, s2|
 *     s1.puts "ping"
 *     s1.shutdown(:WR)
 *     p s2.read          #=> "ping\n"
 *     s2.puts "pong"
 *     s2.close
 *     p s1.read          #=> "pong\n"
 *   }
 *
 */
static VALUE
bsock_shutdown(mrb_state *mrb, mrb_value self)
{
  VALUE howto;
  int how;
  mrb_value fd;

  mrb_get_args(mrb, "o", &howto);

  // TODO constants
  how = mrb_nil_p(howto) ? SHUT_RDWR : mrb_fixnum(howto);
  if (how != SHUT_WR && how != SHUT_RD && how != SHUT_RDWR) {
    mrb_raise(mrb, mrb->eArgumentError_class, "`how' should be either :SHUT_RD, :SHUT_WR, :SHUT_RDWR");
  }

  fd = mrb_iv_get(mrb, self, mrb_intern(mrb, "socket_fd"));
  if (shutdown(mrb_fixnum(fd), how) == -1)
    mrb_sys_fail("shutdown");

  return mrb_fixnum_value(0);
}

/*
 * call-seq:
 *   basicsocket.send(mesg, flags [, dest_sockaddr]) => numbytes_sent
 *
 * send _mesg_ via _basicsocket_.
 *
 * _mesg_ should be a string.
 *
 * _flags_ should be a bitwise OR of Socket::MSG_* constants.
 *
 * _dest_sockaddr_ should be a packed sockaddr string or an addrinfo.
 *
 *   TCPSocket.open("localhost", 80) {|s|
 *     s.send "GET / HTTP/1.0\r\n\r\n", 0
 *     p s.read
 *   }
 */
#include <stdio.h>
VALUE
rsock_bsock_send(mrb_state *mrb, mrb_value self)
{
  struct rsock_send_arg arg;
  VALUE flags, to;
  mrb_value fd;

  mrb_get_args(mrb, "ooo", &arg.mesg, &flags, &to);

  if ( mrb_nil_p(arg.mesg) ) {
    mrb_raise(mrb, mrb->eArgumentError_class, "`mesg' is nil");
  }
  if ( mrb_nil_p(flags) ) {
    mrb_raise(mrb, mrb->eArgumentError_class, "`flags' is nil");
  }

  fd = mrb_iv_get(mrb, self, mrb_intern(mrb, "socket_fd"));
  arg.fd = mrb_fixnum(fd);
  arg.flags = mrb_fixnum(flags);

  arg.mesg = mrb_string_value(mrb, &arg.mesg);
  /* TODO udp */
//  if (!mrb_nil_p(to)) {
//        SockAddrStringValue(to);
//        to = rb_str_new4(to);
//        arg.to = (struct sockaddr *)RSTRING_PTR(to);
//        arg.tolen = (socklen_t)RSTRING_LENINT(to);
//    n = rsock_sendto_blocking((void *)&arg);
//  }
//  else {
//    n = rsock_send_blocking((void *)&arg);
//  }
  return rsock_send_blocking(&arg);
}

/*
 * call-seq:
 *   basicsocket.recv(maxlen) => mesg
 *   basicsocket.recv(maxlen, flags) => mesg
 *
 * Receives a message.
 *
 * _maxlen_ is the maximum number of bytes to receive.
 *
 * _flags_ should be a bitwise OR of Socket::MSG_* constants.
 *
 *   UNIXSocket.pair {|s1, s2|
 *     s1.puts "Hello World"
 *     p s2.recv(4)                     #=> "Hell"
 *     p s2.recv(4, Socket::MSG_PEEK)   #=> "o Wo"
 *     p s2.recv(4)                     #=> "o Wo"
 *     p s2.recv(10)                    #=> "rld\n"
 *   }
 */
static VALUE
bsock_recv(mrb_state *mrb, mrb_value self)
{
    return rsock_s_recvfrom(mrb, self, RECV_RECV);
}

static VALUE
rsock_bsock_read(mrb_state *mrb, mrb_value self)
{
  return rsock_s_read(mrb, self);
}

static VALUE
rsock_bsock_write(mrb_state *mrb, mrb_value self)
{
  struct rsock_send_arg arg;
  mrb_value fd;

  mrb_get_args(mrb, "o", &arg.mesg);

  if ( mrb_nil_p(arg.mesg) ) {
    mrb_raise(mrb, mrb->eArgumentError_class, "`string' is nil ");
  }

  fd = mrb_iv_get(mrb, self, mrb_intern(mrb, "socket_fd"));
  arg.fd = mrb_fixnum(fd);
  arg.flags = 0;
  arg.mesg = mrb_string_value(mrb, &arg.mesg);

  return rsock_send_blocking(&arg);
}

/*
 * BasicSocket is the super class for the all socket classes.
 */
void
rsock_init_basicsocket(mrb_state *mrb)
{
  struct RClass *cBasicSocket;

  cBasicSocket = mrb_define_class(mrb, "BasicSocket", mrb->object_class);

/* TODO undef initialize */
//  mrb_undef_method(mrb, cBasicSocket, "initialize");

  mrb_define_method(mrb, cBasicSocket, "close",         bsock_close,            ARGS_NONE());
  mrb_define_method(mrb, cBasicSocket, "shutdown",      bsock_shutdown,         ARGS_OPT(1));
  mrb_define_method(mrb, cBasicSocket, "send",          rsock_bsock_send,       ARGS_REQ(2) | ARGS_OPT(1));
  mrb_define_method(mrb, cBasicSocket, "recv",          bsock_recv,             ARGS_REQ(1) | ARGS_OPT(1));
  mrb_define_method(mrb, cBasicSocket, "read",          rsock_bsock_read,       ARGS_OPT(2));
  mrb_define_method(mrb, cBasicSocket, "write",         rsock_bsock_write,      ARGS_REQ(1));
}
