#include "mruby.h"

#ifndef INCLUDE_REGEXP
#include <stdarg.h>
#include <string.h>
#include "mstring.h"
#include "mnumeric.h"
#include "mrange.h"
#include <ctype.h>
#include "marray.h"
#include "mclass.h"
#include "variable.h"
#include <stdio.h>


#ifndef FALSE
#define FALSE   0
#endif

#ifndef TRUE
#define TRUE    1
#endif

const char ruby_digitmap[] = "0123456789abcdefghijklmnopqrstuvwxyz";

mrb_value mrb_fs;
struct RVarmap *ruby_dyna_vars;
struct SCOPE *ruby_scope;
int mrb_block_given_p();
const unsigned char *re_mbctab;

static const unsigned char mbctab_ascii[] = {
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
};
const unsigned char *re_mbctab = mbctab_ascii;
#define ismbchar(c) re_mbctab[(unsigned char)(c)]
#define mbclen(c)   (re_mbctab[(unsigned char)(c)]+1)


static inline void
str_mod_check(mrb_state *mrb, mrb_value str, char *p, mrb_int len)
{
  struct RString *s = mrb_str_ptr(str);

  if (s->buf != p || s->len != len) {
    mrb_raise(mrb, mrb->eRuntimeError_class, "string modified");
  }
}

mrb_value
mrb_str_buf_new(mrb_state *mrb, size_t capa)
{
  struct RString *s;

  s = mrb_obj_alloc(mrb, MRB_TT_STRING, mrb->string_class);

  if (capa < STR_BUF_MIN_SIZE) {
    capa = STR_BUF_MIN_SIZE;
  }
  s->len = 0;
  s->capa = capa;
  s->buf = mrb_malloc(mrb, capa+1);
  s->buf[0] = '\0';

  return mrb_obj_value(s);
}

mrb_value
str_buf_cat(mrb_state *mrb, mrb_value str, const char *ptr, size_t len)
{
  struct RString *s1 = mrb_str_ptr(str);
  size_t total = s1->len + len;

  if (s1->capa <= total) {
    if (s1->capa != 0) {
      s1->buf = mrb_realloc(mrb, s1->buf, total+1);
    }
    else {
      s1->buf = mrb_malloc(mrb, total+1); //mrb_realloc(mrb, s1->buf, len);
    }
    s1->capa = total;
  }
  memcpy(s1->buf+s1->len, ptr, len);
  s1->len = total;
  s1->buf[total] = 0;
  return str;
}

mrb_value
mrb_str_buf_cat(mrb_state *mrb, mrb_value str, const char *ptr, size_t len)
{
  if (len == 0) return str;
  if (len < 0) {
    mrb_raise(mrb, mrb->eArgumentError_class, "negative string size (or size too big)");
  }
  return str_buf_cat(mrb, str, ptr, len);
}

/*
 *  call-seq:
 *     String.new(str="")   => new_str
 *
 *  Returns a new string object containing a copy of <i>str</i>.
 */

mrb_value
mrb_str_new(mrb_state *mrb, const char *p, size_t len)
{
  struct RString *s;

  if (len == 0) {
    return mrb_str_buf_new(mrb, len);
  }
  s = mrb_obj_alloc(mrb, MRB_TT_STRING, mrb->string_class);
  s->buf = mrb_malloc(mrb, len+1);
  if (p) {
    memcpy(s->buf, p, len);
  }
  s->len = len;
  s->capa = len;
  s->buf[len] ='\0';
  return mrb_obj_value(s);
}

/* ptr==0 is error */
mrb_value
mrb_str_new2(mrb_state *mrb, const char *ptr)
{
  if (!ptr) {
    mrb_raise(mrb, mrb->eArgumentError_class, "NULL pointer given");
  }
  return mrb_str_new(mrb, ptr, strlen(ptr));
}

/*
 *  call-seq: (Caution! NULL string)
 *     String.new(str="")   => new_str
 *
 *  Returns a new string object containing a copy of <i>str</i>.
 */

mrb_value
mrb_str_new_cstr(mrb_state *mrb, const char *p)
{
  struct RString *s;
  size_t len = strlen(p);

  s = mrb_obj_alloc(mrb, MRB_TT_STRING, mrb->string_class);
  s->buf = mrb_malloc(mrb, len+1);
  memcpy(s->buf, p, len);
  s->buf[len] = 0;
  s->len = len;
  s->capa = len;

  return mrb_obj_value(s);
}

/*
 *  call-seq: (Caution! string literal)
 *     String.new(str="")   => new_str
 *
 *  Returns a new string object containing a copy of <i>str</i>.
 */

mrb_value
mrb_str_literal(mrb_state *mrb, mrb_value lit)
{
  struct RString *s = mrb_str_ptr(lit);

  return mrb_str_new(mrb, s->buf, s->len);
}

/*
 *  call-seq:
 *     char* str = String("abcd"), len=strlen("abcd")
 *
 *  Returns a new string object containing a copy of <i>str</i>.
 */
const char*
mrb_str_body(mrb_value str, int *len_p)
{
  struct RString *s = mrb_str_ptr(str);

  *len_p = s->len;
  return s->buf;
}

/*
 *  call-seq: (Caution! String("abcd") change)
 *     String("abcdefg") = String("abcd") + String("efg")
 *
 *  Returns a new string object containing a copy of <i>str</i>.
 */
void
mrb_str_concat(mrb_state *mrb, mrb_value self, mrb_value other)
{
  struct RString *s1 = mrb_str_ptr(self);
  struct RString *s2 = mrb_str_ptr(other);
  size_t len = s1->len + s2->len;

  if (s1->capa < len) {
    s1->capa = len;
    s1->buf = mrb_realloc(mrb, s1->buf, len+1);
  }
  memcpy(s1->buf+s1->len, s2->buf, s2->len);
  s1->len = len;
  s1->buf[len] = 0;
}

/*
 *  call-seq: (Caution! String("abcd") remain)
 *     String("abcdefg") = String("abcd") + String("efg")
 *
 *  Returns a new string object containing a copy of <i>str</i>.
 */
mrb_value
mrb_str_plus(mrb_state *mrb, mrb_value a, mrb_value b)
{
  struct RString *s = mrb_str_ptr(a);
  struct RString *s2 = mrb_str_ptr(b);
  struct RString *t;
  mrb_value r;

  r = mrb_str_new(mrb, 0, s->len + s2->len);
  t = mrb_str_ptr(r);
  memcpy(t->buf, s->buf, s->len);
  memcpy(t->buf + s->len, s2->buf, s2->len);

  return r;
}

/* 15.2.10.5.2  */

/*
 *  call-seq: (Caution! String("abcd") remain) for stack_argument
 *     String("abcdefg") = String("abcd") + String("efg")
 *
 *  Returns a new string object containing a copy of <i>str</i>.
 */
static mrb_value
mrb_str_plus_m(mrb_state *mrb, mrb_value self)
{
  struct RString *s = mrb_str_ptr(self);
  struct RString *t;
  const char *p;
  size_t len;
  mrb_value r;

  mrb_get_args(mrb, "s", &p, &len);
  r = mrb_str_new(mrb, 0, s->len + len);
  t = mrb_str_ptr(r);
  memcpy(t->buf, s->buf, s->len);
  memcpy(t->buf + s->len, p, len);

  return r;
}

/*
 *  call-seq:
 *     len = strlen(String("abcd"))
 *
 *  Returns a new string object containing a copy of <i>str</i>.
 */
static mrb_value
mrb_str_bytesize(mrb_state *mrb, mrb_value self)
{
  struct RString *s = mrb_str_ptr(self);

  return mrb_fixnum_value(s->len);
}

/* 15.2.10.5.26 */
/* 15.2.10.5.33 */
/*
 *  call-seq:
 *     len = strlen(String("abcd"))
 *
 *  Returns a new string object containing a copy of <i>str</i>.
 */
mrb_value
mrb_str_size(mrb_state *mrb, mrb_value self)
{
#ifndef MRUBY_UTF8_STRING
  return mrb_str_bytesize(mrb, self);
#else
  /* need to count number of characters in UTF-8 */
  return mrb_str_bytesize(mrb, self);
#endif
}

void
mrb_str_modify(mrb_value str)
{
  /*if (!str_independent(str))
    str_make_independent(str);*/
}

mrb_value
mrb_str_resize(mrb_state *mrb, mrb_value str, size_t len)
{
  struct RString *s;
  if (len < 0) {
    mrb_raise(mrb, mrb->eArgumentError_class, "negative string size (or size too big)");
  }

  s = mrb_str_ptr(str);
  mrb_str_modify(str);
  if (len != s->len) {
    if (s->len < len || s->len - len > 1024) {
      //REALLOC_N(s->buf, char, len+1);
      s->buf = mrb_realloc(mrb, s->buf, len+1);
      /*if (!FL_TEST(str, STR_NOCAPA)) {
        s->capa = len;
      }*/
      s->capa = len;
    }
    s->len = len;
    s->buf[len] = '\0';  /* sentinel */
  }
  return str;
}
/* ---------------------------*/

/* 15.2.10.5.1  */

/*
 *  call-seq:
 *     str * integer   => new_str
 *
 *  Copy---Returns a new <code>String</code> containing <i>integer</i> copies of
 *  the receiver.
 *
 *     "Ho! " * 3   #=> "Ho! Ho! Ho! "
 */
static mrb_value
mrb_str_times(mrb_state *mrb, mrb_value self)
{
  mrb_value str2;
  mrb_int i,len,times;
  struct RString *s2;
  struct RString *s = mrb_str_ptr(self);

  mrb_get_args(mrb, "i", &times);

  if (times < 0) {
    mrb_raise(mrb, mrb->eArgumentError_class, "negative argument");
  }
  if (times && INT32_MAX/times <  s->len) {
    mrb_raise(mrb, mrb->eArgumentError_class, "argument too big");
  }

  //str2 = mrb_str_new5(str,0, len *= s->len);
  len = s->len * times;
  str2 = mrb_str_new(mrb, 0, len);
  s2 = mrb_str_ptr(str2);
  for (i = 0; i < len; i += s->len) {
    memcpy(s2->buf + i, s->buf, s->len);
  }
  s2->buf[s2->len] = '\0';

  //OBJ_INFECT(str2, str);

  return str2;
}
/* -------------------------------------------------------------- */

#define lesser(a,b) (((a)>(b))?(b):(a))

/* ---------------------------*/
/*
 *  call-seq:
 *     mrb_value str1 <=> mrb_value str2   => int
 *                     >  1
 *                     =  0
 *                     <  -1
 */
int
mrb_str_cmp(mrb_state *mrb, mrb_value str1, mrb_value str2)
{
  mrb_int len;
  mrb_int retval;
  struct RString *s1 = mrb_str_ptr(str1);
  struct RString *s2 = mrb_str_ptr(str2);

  len = lesser(s1->len, s2->len);
  retval = memcmp(s1->buf, s2->buf, len);
  if (retval == 0) {
    if (s1->len == s2->len) return 0;
    if (s1->len > s2->len)  return 1;
    return -1;
  }
  if (retval > 0) return 1;
  return -1;
}

/* 15.2.10.5.3  */

/*
 *  call-seq:
 *     str <=> other_str   => -1, 0, +1
 *
 *  Comparison---Returns -1 if <i>other_str</i> is less than, 0 if
 *  <i>other_str</i> is equal to, and +1 if <i>other_str</i> is greater than
 *  <i>str</i>. If the strings are of different lengths, and the strings are
 *  equal when compared up to the shortest length, then the longer string is
 *  considered greater than the shorter one. If the variable <code>$=</code> is
 *  <code>false</code>, the comparison is based on comparing the binary values
 *  of each character in the string. In older versions of Ruby, setting
 *  <code>$=</code> allowed case-insensitive comparisons; this is now deprecated
 *  in favor of using <code>String#casecmp</code>.
 *
 *  <code><=></code> is the basis for the methods <code><</code>,
 *  <code><=</code>, <code>></code>, <code>>=</code>, and <code>between?</code>,
 *  included from module <code>Comparable</code>.  The method
 *  <code>String#==</code> does not use <code>Comparable#==</code>.
 *
 *     "abcdef" <=> "abcde"     #=> 1
 *     "abcdef" <=> "abcdef"    #=> 0
 *     "abcdef" <=> "abcdefg"   #=> -1
 *     "abcdef" <=> "ABCDEF"    #=> 1
 */
static mrb_value
mrb_str_cmp_m(mrb_state *mrb, mrb_value str1)
{
  mrb_value str2;
  mrb_int result;

  mrb_get_args(mrb, "o", &str2);
  if (mrb_type(str2) != MRB_TT_STRING) {
    if (!mrb_respond_to(mrb, str2, mrb_intern(mrb, "to_s"))) {
      return mrb_nil_value();
    }
    else if (!mrb_respond_to(mrb, str2, mrb_intern(mrb, "<=>"))) {
      return mrb_nil_value();
    }
    else
    {
      mrb_value tmp = mrb_funcall(mrb, str2, "<=>", 1, str1);

      if (mrb_nil_p(tmp)) return mrb_nil_value();
      if (!mrb_fixnum(tmp)) {
        return mrb_funcall(mrb, mrb_fixnum_value(0), "-", 1, tmp);
      }
      result = -mrb_fixnum(tmp);
    }
  }
  else {
    result = mrb_str_cmp(mrb, str1, str2);
  }
  return mrb_fixnum_value(result);
}

int
mrb_str_equal(mrb_state *mrb, mrb_value str1, mrb_value str2)
{
  struct RString *s2;
  struct RString *s1 = mrb_str_ptr(str1);

  if (mrb_obj_equal(mrb, str1, str2)) return TRUE;
  if (mrb_type(str2) != MRB_TT_STRING) {
    if (!mrb_respond_to(mrb, str2, mrb_intern(mrb, "to_s"))) {
      return FALSE;
    }
    str2 = mrb_funcall(mrb, str2, "to_s", 0);
    return mrb_equal(mrb, str2, str1);
  }
  s2 = mrb_str_ptr(str2);
  if (s1->len == s2->len &&
      mrb_str_cmp(mrb, str1, str2) == 0) {
    return TRUE;
  }
  return FALSE;
}

/* 15.2.10.5.4  */
/*
 *  call-seq:
 *     str == obj   => true or false
 *
 *  Equality---
 *  If <i>obj</i> is not a <code>String</code>, returns <code>false</code>.
 *  Otherwise, returns <code>false</code> or <code>true</code>
 *
 *   caution:if <i>str</i> <code><=></code> <i>obj</i> returns zero.
 */
static mrb_value
mrb_str_equal_m(mrb_state *mrb, mrb_value str1)
{
  mrb_value str2;

  mrb_get_args(mrb, "o", &str2);
  if (mrb_str_equal(mrb, str1, str2)) return mrb_true_value();
  return mrb_false_value();

}
/* ---------------------------------- */
mrb_value
mrb_str_to_str(mrb_state *mrb, mrb_value str)
{
    return mrb_convert_type(mrb, str, MRB_TT_STRING, "String", "to_s");
}

mrb_value
mrb_string_value(mrb_state *mrb, mrb_value *ptr)
{
  struct RString *ps;
  mrb_value s = *ptr;
  if (mrb_type(s) != MRB_TT_STRING) {
    s = mrb_str_to_str(mrb, s);
    *ptr = s;
  }
  ps = mrb_str_ptr(s);
  if (!ps->buf) {
    /*make_null_str(s);*/
  }
  return s;
}
static mrb_value
mrb_str_match(mrb_state *mrb, mrb_value self/* x */)
{
  mrb_value y;
  mrb_get_args(mrb, "o", &y);
  switch (mrb_type(y)) {
    case MRB_TT_STRING:
      mrb_raise(mrb, mrb->eTypeError_class, "type mismatch: String given");
    case MRB_TT_REGEX:
      //return mrb_reg_match_str(mrb, y, self);
      mrb_raise(mrb, mrb->eTypeError_class, "Regexp Class not supported");
      return mrb_nil_value();
    default:
      return mrb_funcall(mrb, y, "=~", 1, self);
  }
}
/* ---------------------------------- */
mrb_value
mrb_str_substr(mrb_state *mrb, mrb_value str, mrb_int beg, int len)
{
  mrb_value str2;
  struct RString *s = mrb_str_ptr(str);
  struct RString *s2;

  if (len < 0) return mrb_nil_value();
  if (beg > (int)(s->len)) return mrb_nil_value();
  if (beg < 0) {
    beg += s->len;
    if (beg < 0) return mrb_nil_value();
  }
  if (beg + len > s->len) {
    len = s->len - beg;
  }
  if (len < 0) {
    len = 0;
  }
  if (len == 0) {
    //str2 = mrb_str_new5(str,0,0);
    str2 = mrb_str_new(mrb, 0, 0);
  }
  else if (len > sizeof(struct RString)/2 &&
           beg + len == s->len /*&& !FL_TEST(str, STR_ASSOC)*/) {
    //str2 = mrb_str_new4(mrb, str);
    //str2 = str_new3(mrb_obj_class(str2), str2);
    str2 = mrb_str_new(mrb, s->buf, s->len);
    s2 = mrb_str_ptr(str2);

    s2->buf += s2->len - len;
    s2->len = len;
  }
  else {
    //str2 = mrb_str_new5(str, s->buf+beg, len);
    str2 = mrb_str_new(mrb, s->buf+beg, len);
  }
  /* OBJ_INFECT(str2, str); */

  return str2;
}

/* --- 1-8-7parse.c --> */
void
mrb_lastline_set(mrb_value val)
{
}

void
mrb_backref_set(mrb_value val)
{
}
/* --- 1-8-7parse.c --< */

static inline long
mrb_memsearch_ss(const unsigned char *xs, long m, const unsigned char *ys, long n)
{
  const unsigned char *x = xs, *xe = xs + m;
  const unsigned char *y = ys, *ye = ys + n;
//2011/06/30 #define SIZEOF_VALUE 4
#ifndef VALUE_MAX
# if SIZEOF_VALUE == 8
#  define VALUE_MAX 0xFFFFFFFFFFFFFFFFULL
# elif SIZEOF_VALUE == 4
#  define VALUE_MAX 0xFFFFFFFFUL
# elif SIZEOF_LONG == SIZEOF_VOIDP
#  define SIZEOF_VALUE 4
#  define VALUE_MAX 0xFFFFFFFFUL
# endif
#endif
  int hx, hy, mask = VALUE_MAX >> ((SIZEOF_VALUE - m) * CHAR_BIT);

  if (m > SIZEOF_VALUE)
    mrb_bug("!!too long pattern string!!");

  /* Prepare hash value */
  for (hx = *x++, hy = *y++; x < xe; ++x, ++y) {
    hx <<= CHAR_BIT;
    hy <<= CHAR_BIT;
    hx |= *x;
    hy |= *y;
  }
  /* Searching */
  while (hx != hy) {
    if (y == ye)
        return -1;
    hy <<= CHAR_BIT;
    hy |= *y;
    hy &= mask;
    y++;
  }
  return y - ys - m;
}

static inline long
mrb_memsearch_qs(const unsigned char *xs, long m, const unsigned char *ys, long n)
{
  const unsigned char *x = xs, *xe = xs + m;
  const unsigned char *y = ys;
  int i, qstable[256];

  /* Preprocessing */
  for (i = 0; i < 256; ++i)
    qstable[i] = m + 1;
  for (; x < xe; ++x)
    qstable[*x] = xe - x;
  /* Searching */
  for (; y + m <= ys + n; y += *(qstable + y[m])) {
    if (*xs == *y && memcmp(xs, y, m) == 0)
        return y - ys;
  }
  return -1;
}
int
mrb_memsearch(const void *x0, int m, const void *y0, int n)
{
  const unsigned char *x = x0, *y = y0;

  if (m > n) return -1;
  else if (m == n) {
    return memcmp(x0, y0, m) == 0 ? 0 : -1;
  }
  else if (m < 1) {
    return 0;
  }
  else if (m == 1) {
    const unsigned char *ys = y, *ye = ys + n;
    for (; y < ye; ++y) {
      if (*x == *y)
        return y - ys;
    }
    return -1;
  }
  else if (m <= SIZEOF_VALUE) {
    return mrb_memsearch_ss(x0, m, y0, n);
  }
  else {
    return mrb_memsearch_qs(x0, m, y0, n);
  }
}

static mrb_int
mrb_str_index(mrb_value str, mrb_value sub, mrb_int offset)
{
  mrb_int pos;
  struct RString *s = mrb_str_ptr(str);
  struct RString *psub = mrb_str_ptr(sub);

  if (offset < 0) {
    offset += s->len;
    if (offset < 0) return -1;
  }
  if (s->len - offset < psub->len) return -1;
  if (psub->len == 0) return offset;
  pos = mrb_memsearch(psub->buf, psub->len,
           s->buf+offset, s->len-offset);
  if (pos < 0) return pos;
  return pos + offset;
}

mrb_value
mrb_str_dup(mrb_state *mrb, mrb_value str)
{
    //mrb_value dup = str_alloc(mrb_obj_class(str));
    //mrb_str_replace(dup, str);
    //return dup;
  struct RString *s = mrb_str_ptr(str);
  struct RString *dup;

  dup = mrb_obj_alloc(mrb, MRB_TT_STRING, mrb->string_class);
  dup->buf = mrb_malloc(mrb, s->capa+1);
  if (s->buf) {
    memcpy(dup->buf, s->buf, s->len);
    dup->buf[s->len]=0;
  }
  dup->len = s->len;
  dup->capa = s->capa;
  return mrb_obj_value(dup);
}

static mrb_value
mrb_str_aref(mrb_state *mrb, mrb_value str, mrb_value indx)
{
  long idx;
  struct RString *s = mrb_str_ptr(str);

  switch (mrb_type(indx)) {
    case MRB_TT_FIXNUM:
      idx = mrb_fixnum(indx);

num_index:
      str = mrb_str_substr(mrb, str, idx, 1);
      if (!mrb_nil_p(str) && RSTRING_LEN(str) == 0) return mrb_nil_value();
      return str;

    case MRB_TT_REGEX:
      //return mrb_str_subpat(mrb, str, indx, 0);
      mrb_raise(mrb, mrb->eTypeError_class, "Regexp Class not supported");
      return mrb_nil_value();

    case MRB_TT_STRING:
      if (mrb_str_index(str, indx, 0) != -1)
        return mrb_str_dup(mrb, indx);
      return mrb_nil_value();

    default:
      /* check if indx is Range */
      {
        mrb_int beg, len;
        mrb_value tmp;

        switch (mrb_range_beg_len(mrb, indx, &beg, &len, s->len, 0)) {
          case 0/*FLASE*/:
            break;
          case 2/*OTHER*/:
            return mrb_nil_value();
          default:
            tmp = mrb_str_substr(mrb, str, beg, len);
            /*OBJ_INFECT(tmp, indx);*/
            return tmp;
        }
      }
      idx = mrb_fixnum(indx);
      goto num_index;
    }
    return mrb_nil_value();    /* not reached */
}

/* 15.2.10.5.6  */
/* 15.2.10.5.34 */
/*
 *  call-seq:
 *     str[fixnum]                 => fixnum or nil
 *     str[fixnum, fixnum]         => new_str or nil
 *     str[range]                  => new_str or nil
 *     str[regexp]                 => new_str or nil
 *     str[regexp, fixnum]         => new_str or nil
 *     str[other_str]              => new_str or nil
 *     str.slice(fixnum)           => fixnum or nil
 *     str.slice(fixnum, fixnum)   => new_str or nil
 *     str.slice(range)            => new_str or nil
 *     str.slice(regexp)           => new_str or nil
 *     str.slice(regexp, fixnum)   => new_str or nil
 *     str.slice(other_str)        => new_str or nil
 *
 *  Element Reference---If passed a single <code>Fixnum</code>, returns the code
 *  of the character at that position. If passed two <code>Fixnum</code>
 *  objects, returns a substring starting at the offset given by the first, and
 *  a length given by the second. If given a range, a substring containing
 *  characters at offsets given by the range is returned. In all three cases, if
 *  an offset is negative, it is counted from the end of <i>str</i>. Returns
 *  <code>nil</code> if the initial offset falls outside the string, the length
 *  is negative, or the beginning of the range is greater than the end.
 *
 *  If a <code>Regexp</code> is supplied, the matching portion of <i>str</i> is
 *  returned. If a numeric parameter follows the regular expression, that
 *  component of the <code>MatchData</code> is returned instead. If a
 *  <code>String</code> is given, that string is returned if it occurs in
 *  <i>str</i>. In both cases, <code>nil</code> is returned if there is no
 *  match.
 *
 *     a = "hello there"
 *     a[1]                   #=> 101
 *     a[1,3]                 #=> "ell"
 *     a[1..3]                #=> "ell"
 *     a[-3,2]                #=> "er"
 *     a[-4..-2]              #=> "her"
 *     a[12..-1]              #=> nil
 *     a[-2..-4]              #=> ""
 *     a[/[aeiou](.)\1/]      #=> "ell"
 *     a[/[aeiou](.)\1/, 0]   #=> "ell"
 *     a[/[aeiou](.)\1/, 1]   #=> "l"
 *     a[/[aeiou](.)\1/, 2]   #=> nil
 *     a["lo"]                #=> "lo"
 *     a["bye"]               #=> nil
 */
static mrb_value
mrb_str_aref_m(mrb_state *mrb, mrb_value str)
{
  int argc;
  mrb_value *argv;

  mrb_get_args(mrb, "*", &argv, &argc);
  if (argc == 2) {
    if (mrb_type(argv[0]) == MRB_TT_REGEX) {
      //return mrb_str_subpat(mrb, str, argv[0], mrb_fixnum(argv[1]));
      mrb_raise(mrb, mrb->eTypeError_class, "Regexp Class not supported");
      return mrb_nil_value();
    }
    return mrb_str_substr(mrb, str, mrb_fixnum(argv[0]), mrb_fixnum(argv[1]));
  }
  if (argc != 1) {
    mrb_raise(mrb, mrb->eArgumentError_class, "wrong number of arguments (%d for 1)", argc);
  }
  return mrb_str_aref(mrb, str, argv[0]);
}

/*
 *  call-seq:
 *     str.capitalize!   => str or nil
 *
 *  Modifies <i>str</i> by converting the first character to uppercase and the
 *  remainder to lowercase. Returns <code>nil</code> if no changes are made.
 *
 *     a = "hello"
 *     a.capitalize!   #=> "Hello"
 *     a               #=> "Hello"
 *     a.capitalize!   #=> nil
 */
static mrb_value
mrb_str_capitalize_bang(mrb_state *mrb, mrb_value str)
{
  char *s, *send;
  int modify = 0;
  struct RString *ps = mrb_str_ptr(str);

  mrb_str_modify(str);
  if (ps->len == 0 || !ps->buf) return mrb_nil_value();
  s = ps->buf; send = s + ps->len;
  if (ISLOWER(*s)) {
    *s = toupper(*s);
    modify = 1;
  }
  while (++s < send) {
    if (ismbchar(*s)) {
      s+=mbclen(*s) - 1;
    }
    else if (ISUPPER(*s)) {
      *s = tolower(*s);
      modify = 1;
    }
  }
  if (modify) return str;
  return mrb_nil_value();
}

/* 15.2.10.5.7  */
/*
 *  call-seq:
 *     str.capitalize   => new_str
 *
 *  Returns a copy of <i>str</i> with the first character converted to uppercase
 *  and the remainder to lowercase.
 *
 *     "hello".capitalize    #=> "Hello"
 *     "HELLO".capitalize    #=> "Hello"
 *     "123ABC".capitalize   #=> "123abc"
 */
static mrb_value
mrb_str_capitalize(mrb_state *mrb, mrb_value self)
{
  mrb_value str;

  str = mrb_str_dup(mrb, self);
  mrb_str_capitalize_bang(mrb, str);
  return str;
}

/* 15.2.10.5.10  */
/*
 *  call-seq:
 *     str.chomp!(separator=$/)   => str or nil
 *
 *  Modifies <i>str</i> in place as described for <code>String#chomp</code>,
 *  returning <i>str</i>, or <code>nil</code> if no modifications were made.
 */
static mrb_value
mrb_str_chomp_bang(mrb_state *mrb, mrb_value self)
{
  mrb_value rs;
  mrb_int newline;
  char *p;
  mrb_int len, rslen;
  struct RString *s = mrb_str_ptr(self);
  struct RString *prs;

  mrb_get_args(mrb, ";o", &rs);
  if (mrb_nil_p(rs) == 0) {
    len = s->len;
    if (len == 0) return mrb_nil_value();
    p = s->buf;
smart_chomp:
    mrb_str_modify(self);
    if (s->buf[len-1] == '\n') {
      s->len--;
      if (s->len > 0 &&
	  s->buf[s->len-1] == '\r') {
	s->len--;
      }
    }
    else if (s->buf[len-1] == '\r') {
      s->len--;
    }
    else {
      return mrb_nil_value();
    }
    s->buf[s->len] = '\0';
    return self;
  }
  if (mrb_nil_p(rs)) return mrb_nil_value();
  //StringValue(rs);
  mrb_string_value(mrb, &rs);
  len = s->len;
  if (len == 0) return mrb_nil_value();
  p = s->buf;
  prs = mrb_str_ptr(rs);
  rslen = prs->len;
  if (rslen == 0) {
    while (len>0 && p[len-1] == '\n') {
      len--;
      if (len>0 && p[len-1] == '\r')
        len--;
    }
    if (len < s->len) {
      mrb_str_modify(self);
      s->len = len;
      s->buf[len] = '\0';
      return self;
    }
    return mrb_nil_value();
  }
  if (rslen > len) return mrb_nil_value();
  newline = prs->buf[rslen-1];
  if (rslen == 1 && newline == '\n')
    goto smart_chomp;

  if (p[len-1] == newline &&
     (rslen <= 1 ||
     memcmp(prs->buf, p+len-rslen, rslen) == 0)) {
    mrb_str_modify(self);
    s->len -= rslen;
    s->buf[s->len] = '\0';
    return self;
  }
  return mrb_nil_value();
}

/* 15.2.10.5.9  */
/*
 *  call-seq:
 *     str.chomp(separator=$/)   => new_str
 *
 *  Returns a new <code>String</code> with the given record separator removed
 *  from the end of <i>str</i> (if present). If <code>$/</code> has not been
 *  changed from the default Ruby record separator, then <code>chomp</code> also
 *  removes carriage return characters (that is it will remove <code>\n</code>,
 *  <code>\r</code>, and <code>\r\n</code>).
 *
 *     "hello".chomp            #=> "hello"
 *     "hello\n".chomp          #=> "hello"
 *     "hello\r\n".chomp        #=> "hello"
 *     "hello\n\r".chomp        #=> "hello\n"
 *     "hello\r".chomp          #=> "hello"
 *     "hello \n there".chomp   #=> "hello \n there"
 *     "hello".chomp("llo")     #=> "he"
 */
static mrb_value
mrb_str_chomp(mrb_state *mrb, mrb_value self)
{
  mrb_value str;

  str = mrb_str_dup(mrb, self);
  mrb_str_chomp_bang(mrb, str);
  return str;
}

/*
 *  call-seq:
 *     str.chop!   => str or nil
 *
 *  Processes <i>str</i> as for <code>String#chop</code>, returning <i>str</i>,
 *  or <code>nil</code> if <i>str</i> is the empty string.  See also
 *  <code>String#chomp!</code>.
 */
static mrb_value
mrb_str_chop_bang(mrb_state *mrb, mrb_value self)
{
  struct RString *s = mrb_str_ptr(self);

  if (s->len > 0) {
    mrb_str_modify(self);
    s->len--;
    if (s->buf[s->len] == '\n') {
      if (s->len > 0 &&
          s->buf[s->len-1] == '\r') {
        s->len--;
      }
    }
    s->buf[s->len] = '\0';
    return self;
  }
  return mrb_nil_value();
}

/* 15.2.10.5.11 */
/*
 *  call-seq:
 *     str.chop   => new_str
 *
 *  Returns a new <code>String</code> with the last character removed.  If the
 *  string ends with <code>\r\n</code>, both characters are removed. Applying
 *  <code>chop</code> to an empty string returns an empty
 *  string. <code>String#chomp</code> is often a safer alternative, as it leaves
 *  the string unchanged if it doesn't end in a record separator.
 *
 *     "string\r\n".chop   #=> "string"
 *     "string\n\r".chop   #=> "string\n"
 *     "string\n".chop     #=> "string"
 *     "string".chop       #=> "strin"
 *     "x".chop            #=> ""
 */
static mrb_value
mrb_str_chop(mrb_state *mrb, mrb_value self)
{
  mrb_value str;

  str = mrb_str_dup(mrb, self);
  mrb_str_chop_bang(mrb, str);
  return str;
}

/* 15.2.10.5.14 */
/*
 *  call-seq:
 *     str.downcase!   => str or nil
 *
 *  Downcases the contents of <i>str</i>, returning <code>nil</code> if no
 *  changes were made.
 */
static mrb_value
mrb_str_downcase_bang(mrb_state *mrb, mrb_value self)
{
  char *s, *send;
  int modify = 0;
  struct RString *ps = mrb_str_ptr(self);

  mrb_str_modify(self);
  s = ps->buf;
  send = s + ps->len;
  while (s < send) {
    if (ismbchar(*s)) {
      s+=mbclen(*s) - 1;
    }
    else if (ISUPPER(*s)) {
      *s = tolower(*s);
      modify = 1;
    }
    s++;
  }

  if (modify) return self;
  return mrb_nil_value();
}

/* 15.2.10.5.13 */
/*
 *  call-seq:
 *     str.downcase   => new_str
 *
 *  Returns a copy of <i>str</i> with all uppercase letters replaced with their
 *  lowercase counterparts. The operation is locale insensitive---only
 *  characters ``A'' to ``Z'' are affected.
 *
 *     "hEllO".downcase   #=> "hello"
 */
static mrb_value
mrb_str_downcase(mrb_state *mrb, mrb_value self)
{
  mrb_value str;

  str = mrb_str_dup(mrb, self);
  mrb_str_downcase_bang(mrb, str);
  return str;
}

/* 15.2.10.5.15 */
/*
 *  call-seq:
 *     str.each(separator=$/) {|substr| block }        => str
 *     str.each_line(separator=$/) {|substr| block }   => str
 *
 *  Splits <i>str</i> using the supplied parameter as the record separator
 *  (<code>$/</code> by default), passing each substring in turn to the supplied
 *  block. If a zero-length record separator is supplied, the string is split
 *  into paragraphs delimited by multiple successive newlines.
 *
 *     print "Example one\n"
 *     "hello\nworld".each {|s| p s}
 *     print "Example two\n"
 *     "hello\nworld".each('l') {|s| p s}
 *     print "Example three\n"
 *     "hello\n\n\nworld".each('') {|s| p s}
 *
 *  <em>produces:</em>
 *
 *     Example one
 *     "hello\n"
 *     "world"
 *     Example two
 *     "hel"
 *     "l"
 *     "o\nworl"
 *     "d"
 *     Example three
 *     "hello\n\n\n"
 *     "world"
 */
static mrb_value
mrb_str_each_line(mrb_state *mrb, mrb_value str)
{
  mrb_value rs;
  int newline;
  struct RString *ps = mrb_str_ptr(str);
  char *p = ps->buf, *pend = p + ps->len, *s;
  char *ptr = p;
  long len = ps->len, rslen;
  mrb_value line;
  struct RString *prs;
  int n;

  //if (mrb_scan_args(argc, argv, "01", &rs) == 0) {
  n = mrb_get_args(mrb, ";o", &rs);
  /*RETURN_ENUMERATOR(str, argc, argv);*/
  if (n > 0 && mrb_nil_p(rs)) {
    mrb_yield(str);
    return str;
  }
  //StringValue(rs);
  mrb_string_value(mrb, &rs);
  prs = mrb_str_ptr(rs);
  rslen = prs->len;
  if (rslen == 0) {
    newline = '\n';
  }
  else {
    newline = prs->buf[rslen-1];
  }

  for (s = p, p += rslen; p < pend; p++) {
    if (rslen == 0 && *p == '\n') {
      if (*++p != '\n') continue;
      while (*p == '\n') p++;
    }
    if (ps->buf < p && p[-1] == newline &&
       (rslen <= 1 ||
        memcmp(prs->buf, p-rslen, rslen) == 0)) {
      //line = mrb_str_new5(str, s, p - s);
      line = mrb_str_new(mrb, s, p - s);
      /*OBJ_INFECT(line, str);*/
      mrb_yield(line);
      str_mod_check(mrb, str, ptr, len);
      s = p;
    }
  }

  if (s != pend) {
    if (p > pend) p = pend;
    //line = mrb_str_new5(str, s, p - s);
    line = mrb_str_new(mrb, s, p - s);
    /*OBJ_INFECT(line, str);*/
    mrb_yield(line);
  }

  return str;
}

/* 15.2.10.5.16 */
/*
 *  call-seq:
 *     str.empty?   => true or false
 *
 *  Returns <code>true</code> if <i>str</i> has a length of zero.
 *
 *     "hello".empty?   #=> false
 *     "".empty?        #=> true
 */
static mrb_value
mrb_str_empty(mrb_state *mrb, mrb_value self)
{
  struct RString *s = mrb_str_ptr(self);

  if (s->len == 0)
    return mrb_true_value();
  return mrb_false_value();
}

/* 15.2.10.5.17 */
/*
 * call-seq:
 *   str.eql?(other)   => true or false
 *
 * Two strings are equal if the have the same length and content.
 */
static mrb_value
mrb_str_eql(mrb_state *mrb, mrb_value self)
{
  mrb_int rslen;
  const char* prs;
  struct RString *s = mrb_str_ptr(self);

  mrb_get_args(mrb, "s", &prs, &rslen);
  if (/*mrb_type(str2) != MRB_TT_STRING || */s->len != rslen)
    return mrb_false_value();

  if (memcmp(s->buf, prs,
         lesser(s->len, rslen)) == 0)
    return mrb_true_value();

  return mrb_false_value();
}



mrb_int
mrb_str_hash(mrb_state *mrb, mrb_value str)
{
  struct RString *s = mrb_str_ptr(str);
  long len = s->len;
  char *p = s->buf;
  mrb_int key = 0;

  while (len--) {
    key = key*65599 + *p;
    p++;
  }
  key = key + (key>>5);
  return key;
}

/*
 * call-seq:
 *    str.hash   => fixnum
 *
 * Return a hash based on the string's length and content.
 */
static mrb_value
mrb_str_hash_m(mrb_state *mrb, mrb_value self)
{
  mrb_int key = mrb_str_hash(mrb, self);
  return mrb_fixnum_value(key);
}

/*
 *  call-seq:
 *     str.include? other_str   => true or false
 *     str.include? fixnum      => true or false
 *
 *  Returns <code>true</code> if <i>str</i> contains the given string or
 *  character.
 *
 *     "hello".include? "lo"   #=> true
 *     "hello".include? "ol"   #=> false
 *     "hello".include? ?h     #=> true
 */
static mrb_value
mrb_str_include(mrb_state *mrb, mrb_value self)
{
  mrb_int i;
  struct RString *s = mrb_str_ptr(self);
  mrb_value str2;
  mrb_get_args(mrb, "o", &str2);
  if (mrb_type(str2) == MRB_TT_FIXNUM) {
    if (memchr(s->buf, mrb_fixnum(str2), s->len))
      return mrb_true_value();
    return mrb_false_value();
  }

  //StringValue(arg);
  mrb_string_value(mrb, &str2);
  i = mrb_str_index(self, str2, 0);

  if (i == -1) return mrb_false_value();
  return mrb_true_value();
}

/* 15.2.10.5.22 */
/*
 *  call-seq:
 *     str.index(substring [, offset])   => fixnum or nil
 *     str.index(fixnum [, offset])      => fixnum or nil
 *     str.index(regexp [, offset])      => fixnum or nil
 *
 *  Returns the index of the first occurrence of the given
 *  <i>substring</i>,
 *  character (<i>fixnum</i>), or pattern (<i>regexp</i>) in <i>str</i>.
 *  Returns
 *  <code>nil</code> if not found.
 *  If the second parameter is present, it
 *  specifies the position in the string to begin the search.
 *
 *     "hello".index('e')             #=> 1
 *     "hello".index('lo')            #=> 3
 *     "hello".index('a')             #=> nil
 *     "hello".index(101)             #=> 1(101=0x65='e')
 *     "hello".index(/[aeiou]/, -3)   #=> 4
 */
static mrb_value
mrb_str_index_m(mrb_state *mrb, mrb_value self)
{

  mrb_value *argv;
  int argc;

  mrb_value sub;
  //mrb_value initpos;
  mrb_int pos;
  struct RString *s = mrb_str_ptr(self);

  //if (mrb_scan_args(argc, argv, "11", &sub, &initpos) == 2) {
  mrb_get_args(mrb, "*", &argv, &argc);
  if (argc == 2) {
    //    pos = mrb_fixnum(args[1]);
    pos = mrb_fixnum(argv[1]);//changed by yamashita 
    //    sub = args[0];
    sub = argv[0];//changed by yamashita 
  }
  else {
    pos = 0;
    if (argc > 0)// added by yamashita
      sub = argv[0];//changed by yamashita
    else//added by yamashita
      sub = mrb_nil_value();//added by yamashita

  }
  if (pos < 0) {
    pos += s->len;
    if (pos < 0) {
      if (mrb_type(sub) == MRB_TT_REGEX) {
        //mrb_backref_set(mrb_nil_value());
        mrb_raise(mrb, mrb->eTypeError_class, "Regexp Class not supported");
        return mrb_nil_value();
      }
      return mrb_nil_value();
    }
  }

  switch (mrb_type(sub)) {
    case MRB_TT_REGEX:
      //pos = mrb_reg_adjust_startpos(sub, self, pos, 0);
      //pos = mrb_reg_search(sub, self, pos, 0);
      mrb_raise(mrb, mrb->eTypeError_class, "Regexp Class not supported");
      return mrb_nil_value();
      break;

    case MRB_TT_FIXNUM: {
      int c = mrb_fixnum(sub);
      long len = s->len;
      unsigned char *p = (unsigned char*)s->buf;

      for (;pos<len;pos++) {
        if (p[pos] == c) return mrb_fixnum_value(pos);
      }
      return mrb_nil_value();
    }

    default: {
      mrb_value tmp;

      tmp = mrb_check_string_type(mrb, sub);
      if (mrb_nil_p(tmp)) {
        mrb_raise(mrb, mrb->eTypeError_class, "type mismatch: %s given",
           mrb_obj_classname(mrb, sub));
      }
      sub = tmp;
    }
//     fall through
    case MRB_TT_STRING:
      pos = mrb_str_index(self, sub, pos);
      break;
    }

    if (pos == -1) return mrb_nil_value();
    return mrb_fixnum_value(pos);

}

/* 15.2.10.5.24 */
/* 15.2.10.5.28 */
/*
 *  call-seq:
 *     str.replace(other_str)   => str
 *
 *  Replaces the contents and taintedness of <i>str</i> with the corresponding
 *  values in <i>other_str</i>.
 *
 *     s = "hello"         #=> "hello"
 *     s.replace "world"   #=> "world"
 */
static mrb_value
mrb_str_replace(mrb_state *mrb, mrb_value str)
{
  mrb_value str2;
  struct RString *s2;
  struct RString *s = mrb_str_ptr(str);

  mrb_get_args(mrb, "o", &str2);
  if (mrb_obj_equal(mrb, str, str2)) return str;

  //StringValue(str2);
  mrb_string_value(mrb, &str2);
  s2 = mrb_str_ptr(str2);
  /*if (FL_TEST(str2, ELTS_SHARED)) {
    if (str_independent(str)) {
      free(ps->buf);
    }
    RSTRING(str)->len = RSTRING(str2)->len;
    ps->buf = RSTRING(str2)->ptr;
    FL_SET(str, ELTS_SHARED);
    FL_UNSET(str, STR_ASSOC);
    RSTRING(str)->aux.shared = RSTRING(str2)->aux.shared;
  }
  else*/ {
    /*if (str_independent(str)) {
      mrb_str_resize(mrb, str, RSTRING(str2)->len);
      memcpy(ps->buf, RSTRING(str2)->ptr, RSTRING(str2)->len);
      if (!ps->buf) {
        make_null_str(str);
      }
    }
    else */{
      s->buf = s2->buf;
      s->len = s2->len;
      /*str_make_independent(str);*/
    }
    /*if (FL_TEST(str2, STR_ASSOC)) {
      FL_SET(str, STR_ASSOC);
      RSTRING(str)->aux.shared = RSTRING(str2)->aux.shared;
    }*/
  }

  /*OBJ_INFECT(str, str2);*/
  return str;
}

/* 15.2.10.5.23 */
/*
 *  call-seq:
 *     String.new(str="")   => new_str
 *
 *  Returns a new string object containing a copy of <i>str</i>.
 */
static mrb_value
mrb_str_init(mrb_state *mrb, mrb_value self)
{
  //mrb_value orig;
  mrb_value *argv;
  int argc;

  mrb_get_args(mrb, "*", &argv, &argc);
  if (argc == 1)
    mrb_str_replace(mrb, self);
  return self;
}

/*
 *  call-seq:
 *     str.intern   => symbol
 *     str.to_sym   => symbol
 *
 *  Returns the <code>Symbol</code> corresponding to <i>str</i>, creating the
 *  symbol if it did not previously exist. See <code>Symbol#id2name</code>.
 *
 *     "Koala".intern         #=> :Koala
 *     s = 'cat'.to_sym       #=> :cat
 *     s == :cat              #=> true
 *     s = '@cat'.to_sym      #=> :@cat
 *     s == :@cat             #=> true
 *
 *  This can also be used to create symbols that cannot be represented using the
 *  <code>:xxx</code> notation.
 *
 *     'cat and dog'.to_sym   #=> :"cat and dog"
 */
mrb_value
mrb_str_intern(mrb_state *mrb, mrb_value self)
{
  mrb_value str = self;
  mrb_sym id;
  struct RString *s = mrb_str_ptr(str);

  if (!s->buf || s->len == 0) {
    mrb_raise(mrb, mrb->eArgumentError_class, "interning empty string");
  }
  if (strlen(s->buf) != s->len)
    mrb_raise(mrb, mrb->eArgumentError_class, "symbol string may not contain `\\0'");
  /*if (OBJ_TAINTED(str) && mrb_safe_level() >= 1 && !mrb_sym_interned_p(str)) {
    mrb_raise(mrb, mrb_eSecurityError, "Insecure: can't intern tainted string");
  }*/
  id = mrb_intern(mrb, s->buf);
  //return mrb_str_new(mrb, mrb_sym2name(mrb, id), s->len);
  return mrb_symbol_value(id);
}
/* ---------------------------------- */
mrb_value
mrb_obj_as_string(mrb_state *mrb, mrb_value obj)
{
  mrb_value str;

  if (mrb_type(obj) == MRB_TT_STRING) {
    return obj;
  }
  str = mrb_funcall(mrb, obj, "to_s", 0);
  if (mrb_type(str) != MRB_TT_STRING)
    return mrb_any_to_s(mrb, obj);
  /*if (OBJ_TAINTED(obj)) OBJ_TAINT(str);*/
  return str;
}
#define make_null_str(s) do { \
    FL_SET(s, ELTS_SHARED); \
    s->buf = (char *)null_str; \
    s->aux.shared = 0; \
    } while (0)

mrb_value
mrb_check_string_type(mrb_state *mrb, mrb_value str)
{
  struct RString *s = mrb_str_ptr(str);
  str = mrb_check_convert_type(mrb, str, MRB_TT_STRING, "String", "to_s");
  if (!mrb_nil_p(str) && !s->buf) {
    /* make_null_str(str); */
  }
  return str;
}

mrb_value
mrb_reg_quote(mrb_state *mrb, mrb_value str)
{
  char *s, *send, *t;
  mrb_value tmp;
  int c;
  struct RString *ps;
  struct RString *pstmp;

  ps = mrb_str_ptr(str);
  s = ps->buf;
  send = s + ps->len;
  for (; s < send; s++) {
    c = *s;
    if (ismbchar(c)) {
      int n = mbclen(c);

      while (n-- && s < send)
      s++;
      s--;
      continue;
    }
    switch (c) {
      case '[': case ']': case '{': case '}':
      case '(': case ')': case '|': case '-':
      case '*': case '.': case '\\':
      case '?': case '+': case '^': case '$':
      case ' ': case '#':
      case '\t': case '\f': case '\n': case '\r':
        goto meta_found;
    }
  }
  /* before shared  return mrb_str_new3(str); */
  return str;

meta_found:
  tmp = mrb_str_new(mrb, 0, ps->len*2);
  pstmp = mrb_str_ptr(tmp);
  t = pstmp->buf;
  /* copy upto metacharacter */
  memcpy(t, ps->buf, s - ps->buf);
  t += s - ps->buf;

  for (; s < send; s++) {
    c = *s;
    if (ismbchar(c)) {
      int n = mbclen(c);

      while (n-- && s < send)
        *t++ = *s++;
      s--;
      continue;
    }
    switch (c) {
      case '[': case ']': case '{': case '}':
      case '(': case ')': case '|': case '-':
      case '*': case '.': case '\\':
      case '?': case '+': case '^': case '$':
      case '#':
        *t++ = '\\';
        break;
      case ' ':
        *t++ = '\\';
        *t++ = ' ';
        continue;
      case '\t':
        *t++ = '\\';
        *t++ = 't';
        continue;
      case '\n':
        *t++ = '\\';
        *t++ = 'n';
        continue;
      case '\r':
        *t++ = '\\';
        *t++ = 'r';
        continue;
      case '\f':
        *t++ = '\\';
        *t++ = 'f';
        continue;
    }
    *t++ = c;
  }
  mrb_str_resize(mrb, tmp, t - pstmp->buf);
  /*OBJ_INFECT(tmp, str);*/
  return tmp;
}

/* ---------------------------------- */
/* 15.2.10.5.29 */
/*
 *  call-seq:
 *     str.reverse   => new_str
 *
 *  Returns a new string with the characters from <i>str</i> in reverse order.
 *
 *     "stressed".reverse   #=> "desserts"
 */
static mrb_value
mrb_str_reverse(mrb_state *mrb, mrb_value str)
{
  mrb_value obj;
  char *s, *e, *p;
  struct RString *ps = mrb_str_ptr(str);
  struct RString *news;
  if (ps->len <= 1) return mrb_str_dup(mrb, str);

  obj = mrb_str_new(mrb, 0, ps->len);
  news = mrb_str_ptr(obj);
  s = ps->buf; e = s + ps->len - 1;
  p = news->buf;

  while (e >= s) {
    *p++ = *e--;
  }
  /*OBJ_INFECT(obj, str);*/

  return obj;
}

/* 15.2.10.5.30 */
/*
 *  call-seq:
 *     str.reverse!   => str
 *
 *  Reverses <i>str</i> in place.
 */
static mrb_value
mrb_str_reverse_bang(mrb_state *mrb, mrb_value str)
{
  char *s, *e;
  char c;
  struct RString *ps = mrb_str_ptr(str);

  if (ps->len > 1) {
    mrb_str_modify(str);
    s = ps->buf;
    e = s + ps->len - 1;
    while (s < e) {
      c = *s;
      *s++ = *e;
      *e-- = c;
    }
  }
  return str;
}

/*
 *  call-seq:
 *     str.rindex(substring [, fixnum])   => fixnum or nil
 *     str.rindex(fixnum [, fixnum])   => fixnum or nil
 *     str.rindex(regexp [, fixnum])   => fixnum or nil
 *
 *  Returns the index of the last occurrence of the given <i>substring</i>,
 *  character (<i>fixnum</i>), or pattern (<i>regexp</i>) in <i>str</i>. Returns
 *  <code>nil</code> if not found. If the second parameter is present, it
 *  specifies the position in the string to end the search---characters beyond
 *  this point will not be considered.
 *
 *     "hello".rindex('e')             #=> 1
 *     "hello".rindex('l')             #=> 3
 *     "hello".rindex('a')             #=> nil
 *     "hello".rindex(101)             #=> 1
 *     "hello".rindex(/[aeiou]/, -2)   #=> 1
 */
static mrb_int
mrb_str_rindex(mrb_state *mrb, mrb_value str, mrb_value sub, mrb_int pos)
{
  char *s, *sbeg, *t;
  struct RString *ps = mrb_str_ptr(str);
  struct RString *psub = mrb_str_ptr(sub);
  long len = psub->len;

  /* substring longer than string */
  if (ps->len < len) return -1;
  if (ps->len - pos < len) {
    pos = ps->len - len;
  }
  sbeg = ps->buf;
  s = ps->buf + pos;
  t = psub->buf;
  if (len) {
    while (sbeg <= s) {
      if (memcmp(s, t, len) == 0) {
        return s - ps->buf;
      }
      s--;
    }
    return -1;
  }
  else {
    return pos;
  }
}

/* 15.2.10.5.31 */
/*
 *  call-seq:
 *     str.rindex(substring [, fixnum])   => fixnum or nil
 *     str.rindex(fixnum [, fixnum])   => fixnum or nil
 *     str.rindex(regexp [, fixnum])   => fixnum or nil
 *
 *  Returns the index of the last occurrence of the given <i>substring</i>,
 *  character (<i>fixnum</i>), or pattern (<i>regexp</i>) in <i>str</i>. Returns
 *  <code>nil</code> if not found. If the second parameter is present, it
 *  specifies the position in the string to end the search---characters beyond
 *  this point will not be considered.
 *
 *     "hello".rindex('e')             #=> 1
 *     "hello".rindex('l')             #=> 3
 *     "hello".rindex('a')             #=> nil
 *     "hello".rindex(101)             #=> 1
 *     "hello".rindex(/[aeiou]/, -2)   #=> 1
 */
static mrb_value
mrb_str_rindex_m(mrb_state *mrb, mrb_value str)
{
  mrb_value *argv;
  int argc;
  mrb_value sub;
  mrb_value vpos; // added by yamashita
  //mrb_value position;
  mrb_int pos;
  struct RString *ps = mrb_str_ptr(str);
  struct RString *psub;

  //if (mrb_scan_args(argc, *argv, "11", &sub, &position) == 2) {
  mrb_get_args(mrb, "*", &argv, &argc);
  if (argc == 2) {
    sub = argv[0];//added by yamashita
    vpos = argv[1];//added by yamashita
    pos = mrb_fixnum(vpos);//changed by yamashit
    if (pos < 0) {
      pos += ps->len;
      if (pos < 0) {
        if (mrb_type(sub) == MRB_TT_REGEX) {
          //mrb_backref_set(mrb_nil_value());
          mrb_raise(mrb, mrb->eTypeError_class, "Regexp Class not supported");
          return mrb_nil_value();
        }
        return mrb_nil_value();
      }
    }
    if (pos > ps->len) pos = ps->len;
  }
  else {
    pos = ps->len;
  }
  sub = argv[0];//changed by yamashita

  psub = mrb_str_ptr(sub);
  switch (mrb_type(sub)) {
    case MRB_TT_REGEX:
      //pr = mrb_regex_ptr(sub);
      //if (pr->len) {
      //  pos = mrb_reg_adjust_startpos(sub, str, pos, 1);
      //  pos = mrb_reg_search(sub, str, pos, 1);
      //}
      //if (pos >= 0) return mrb_fixnum_value(pos);
      mrb_raise(mrb, mrb->eTypeError_class, "Regexp Class not supported");
      return mrb_nil_value();
      break;

    default: {
      mrb_value tmp;

      tmp = mrb_check_string_type(mrb, sub);
      if (mrb_type(tmp)) {
        mrb_raise(mrb, mrb->eTypeError_class, "type mismatch: %s given",
                 mrb_obj_classname(mrb, sub));
      }
      sub = tmp;
    }
     /* fall through */
    case MRB_TT_STRING:
      pos = mrb_str_rindex(mrb, str, sub, pos);
      if (pos >= 0) return mrb_fixnum_value(pos);
      break;

    case MRB_TT_FIXNUM: {
      int c = mrb_fixnum(sub);
      unsigned char *p = (unsigned char*)ps->buf + pos;
      unsigned char *pbeg = (unsigned char*)ps->buf;

      if (pos == ps->len) {
        if (pos == 0) return mrb_nil_value();
        --p;
      }
      while (pbeg <= p) {
        if (*p == c) return mrb_fixnum_value((char*)p - ps->buf);
        p--;
      }
      return mrb_nil_value();
    }
  } /* end of switch (TYPE(sub)) */
  return mrb_nil_value();
}

/*
 *  call-seq:
 *     str.split(pattern=$;, [limit])   => anArray
 *
 *  Divides <i>str</i> into substrings based on a delimiter, returning an array
 *  of these substrings.
 *
 *  If <i>pattern</i> is a <code>String</code>, then its contents are used as
 *  the delimiter when splitting <i>str</i>. If <i>pattern</i> is a single
 *  space, <i>str</i> is split on whitespace, with leading whitespace and runs
 *  of contiguous whitespace characters ignored.
 *
 *  If <i>pattern</i> is a <code>Regexp</code>, <i>str</i> is divided where the
 *  pattern matches. Whenever the pattern matches a zero-length string,
 *  <i>str</i> is split into individual characters.
 *
 *  If <i>pattern</i> is omitted, the value of <code>$;</code> is used.  If
 *  <code>$;</code> is <code>nil</code> (which is the default), <i>str</i> is
 *  split on whitespace as if ` ' were specified.
 *
 *  If the <i>limit</i> parameter is omitted, trailing null fields are
 *  suppressed. If <i>limit</i> is a positive number, at most that number of
 *  fields will be returned (if <i>limit</i> is <code>1</code>, the entire
 *  string is returned as the only entry in an array). If negative, there is no
 *  limit to the number of fields returned, and trailing null fields are not
 *  suppressed.
 *
 *     " now's  the time".split        #=> ["now's", "the", "time"]
 *     " now's  the time".split(' ')   #=> ["now's", "the", "time"]
 *     " now's  the time".split(/ /)   #=> ["", "now's", "", "the", "time"]
 *     "1, 2.34,56, 7".split(%r{,\s*}) #=> ["1", "2.34", "56", "7"]
 *     "hello".split(//)               #=> ["h", "e", "l", "l", "o"]
 *     "hello".split(//, 3)            #=> ["h", "e", "llo"]
 *     "hi mom".split(%r{\s*})         #=> ["h", "i", "m", "o", "m"]
 *
 *     "mellow yellow".split("ello")   #=> ["m", "w y", "w"]
 *     "1,2,,3,4,,".split(',')         #=> ["1", "2", "", "3", "4"]
 *     "1,2,,3,4,,".split(',', 4)      #=> ["1", "2", "", "3,4,,"]
 *     "1,2,,3,4,,".split(',', -4)     #=> ["1", "2", "", "3", "4", "", ""]
 */

//static mrb_value
//mrb_str_split_m(int argc, mrb_value *argv, mrb_value str)
static mrb_value
mrb_str_split_m(mrb_state *mrb, mrb_value str)
{
  mrb_value *argv;
  int argc;
  mrb_value spat;
  mrb_value limit;
  int awk_split = 0/*Qfalse*/;
  long beg, end, i = 0;
  int lim = 0;
  mrb_value result, tmp;
  struct RString *ps = mrb_str_ptr(str);
  struct RString *pspat;
  struct RArray *pary;

  //if (mrb_scan_args(argc, argv, "02", &spat, &limit) == 2) {
  argv[0] = mrb_nil_value();
  argv[1] = mrb_nil_value();
  mrb_get_args(mrb, "*", &argv, &argc);
  spat  = argv[0];
  limit = argv[1];
  if (argc == 2) {
    lim = mrb_fixnum(limit);
    if (lim <= 0) limit = mrb_nil_value();
    else if (lim == 1) {
      if (ps->len == 0)
        return mrb_ary_new_capa(mrb, 0); /*mrb_ary_new2(0);*/
      return mrb_ary_new_from_values(mrb, &str, 1);/*mrb_ary_new3(1, str);*/
    }
    i = 1;
  }

  //if (mrb_nil_p(spat)) {
  if (argc == 0) {
    if (!mrb_nil_p(mrb_fs)) {
      spat = mrb_fs;
      goto fs_set;
    }
    awk_split = 1/*Qtrue*/;
  }
  else {
fs_set:
    pspat = mrb_str_ptr(spat);
    if (mrb_type(spat) == MRB_TT_STRING && pspat->len == 1) {
      if (pspat->buf[0] == ' ') {
        awk_split = 1/*Qtrue*/;
      }
      else {
        mrb_raise(mrb, mrb->eTypeError_class, "Regexp Class not supported");
        return mrb_nil_value();
      }
    }
    else {
        mrb_raise(mrb, mrb->eTypeError_class, "Regexp Class not supported");
        return mrb_nil_value();
    }
  }

  result = mrb_ary_new(mrb);
  beg = 0;
  if (awk_split) {
    char *ptr = ps->buf;
    long len = ps->len;
    char *eptr = ptr + len;
    int skip = 1;

    for (end = beg = 0; ptr<eptr; ptr++) {
      if (skip) {
        if (ISSPACE(*ptr)) {
          beg++;
        }
        else {
          end = beg+1;
          skip = 0;
          if (!mrb_nil_p(limit) && lim <= i) break;
        }
      }
      else {
        if (ISSPACE(*ptr)) {
          mrb_ary_push(mrb, result, mrb_str_substr(mrb, str, beg, end-beg));
          skip = 1;
          beg = end + 1;
          if (!mrb_nil_p(limit)) ++i;
        }
        else {
          end++;
        }
      }
    }
  }
  else {
    mrb_raise(mrb, mrb->eTypeError_class, "Regexp Class not supported");
    return mrb_nil_value();
  }
  if (ps->len > 0 && (!mrb_nil_p(limit) || ps->len > beg || lim < 0)) {
    if (ps->len == beg)
      //tmp = mrb_str_new5(str, 0, 0);
      tmp = mrb_str_new(mrb, 0, 0);
    else
      tmp = mrb_str_substr(mrb, str, beg, ps->len-beg);
    mrb_ary_push(mrb, result, tmp);
  }
  if (mrb_nil_p(limit) && lim == 0) {
    pary = mrb_ary_ptr(result);
    ps = mrb_str_ptr(pary->buf[pary->len-1]);
    while (pary->len > 0 && ps->len == 0) {
      mrb_ary_pop(mrb, result);
      pary = mrb_ary_ptr(result);
      ps = mrb_str_ptr(pary->buf[pary->len-1]);
    }
  }

  return result;
}

mrb_value *
mrb_svar(mrb_int cnt)
{
  struct RVarmap *vars = ruby_dyna_vars;
  mrb_sym id;

  if (!ruby_scope->local_tbl) return NULL;
  if (cnt >= ruby_scope->local_tbl[0]) return NULL;
  id = ruby_scope->local_tbl[cnt+1];
  while (vars) {
    if (vars->id == id) return &vars->val;
    vars = vars->next;
  }
  if (ruby_scope->local_vars == 0) return NULL;
  return &ruby_scope->local_vars[cnt];
}

/*
 *  call-seq:
 *     split([pattern [, limit]])    => array
 *
 *  Equivalent to <code>$_.split(<i>pattern</i>, <i>limit</i>)</code>.
 *  See <code>String#split</code>.
 */
static mrb_value
mrb_str_f_split(mrb_state *mrb, mrb_value self)
{
  //return mrb_str_split_m(argc, argv, uscore_get());
  return mrb_str_split_m(mrb, self);
}

int
mrb_block_given_p()
{
  /*if (ruby_frame->iter == ITER_CUR && ruby_block)
    return 1;*//*Qtrue*/
  return 0/*Qfalse*/;
}


mrb_value
mrb_cstr_to_inum(mrb_state *mrb, const char *str, int base, int badcheck)
{
  #define BDIGIT unsigned int
  #define BDIGIT_DBL unsigned long

//  const char *s = str;
  char *end;
  char sign = 1;
//  char nondigit = 0;
  int c;
//  BDIGIT_DBL num;
  long len;
//  long blen = 1;
//  long i;
//  mrb_value z;
//  BDIGIT *zds;

#define conv_digit(c) \
    (!ISASCII(c) ? -1 : \
     isdigit(c) ? ((c) - '0') : \
     islower(c) ? ((c) - 'a' + 10) : \
     isupper(c) ? ((c) - 'A' + 10) : \
     -1)

  if (!str) {
    if (badcheck) goto bad;
    return mrb_fixnum_value(0);
  }
  if (badcheck) {
    while (ISSPACE(*str)) str++;
  }
  else {
    while (ISSPACE(*str) || *str == '_') str++;
  }

  if (str[0] == '+') {
    str++;
  }
  else if (str[0] == '-') {
    str++;
    sign = 0;
  }
  if (str[0] == '+' || str[0] == '-') {
    if (badcheck) goto bad;
    return mrb_fixnum_value(0);
  }
  if (base <= 0) {
    if (str[0] == '0') {
      switch (str[1]) {
        case 'x': case 'X':
          base = 16;
          break;
        case 'b': case 'B':
          base = 2;
          break;
        case 'o': case 'O':
          base = 8;
          break;
        case 'd': case 'D':
          base = 10;
          break;
        default:
          base = 8;
      }
    }
    else if (base < -1) {
      base = -base;
    }
    else {
      base = 10;
    }
  }
  switch (base) {
    case 2:
      len = 1;
      if (str[0] == '0' && (str[1] == 'b'||str[1] == 'B')) {
        str += 2;
      }
      break;
    case 3:
      len = 2;
      break;
    case 8:
      if (str[0] == '0' && (str[1] == 'o'||str[1] == 'O')) {
        str += 2;
      }
    case 4: case 5: case 6: case 7:
      len = 3;
      break;
    case 10:
      if (str[0] == '0' && (str[1] == 'd'||str[1] == 'D')) {
        str += 2;
      }
    case 9: case 11: case 12: case 13: case 14: case 15:
      len = 4;
      break;
    case 16:
      len = 4;
      if (str[0] == '0' && (str[1] == 'x'||str[1] == 'X')) {
        str += 2;
      }
      break;
    default:
      if (base < 2 || 36 < base) {
        mrb_raise(mrb, mrb->eArgumentError_class, "illegal radix %d", base);
      }
      if (base <= 32) {
        len = 5;
      }
      else {
        len = 6;
      }
      break;
  } /* end of switch (base) { */
  if (*str == '0') {    /* squeeze preceeding 0s */
    int us = 0;
    while ((c = *++str) == '0' || c == '_') {
      if (c == '_') {
        if (++us >= 2)
          break;
      }
      else
        us = 0;
    }
    if (!(c = *str) || ISSPACE(c)) --str;
  }
  c = *str;
  c = conv_digit(c);
  if (c < 0 || c >= base) {
    if (badcheck) goto bad;
    return mrb_fixnum_value(0);
  }
  len *= strlen(str)*sizeof(char);

  //if (len <= (sizeof(mrb_value)*CHAR_BIT)) {
    unsigned long val = strtoul((char*)str, &end, base);

    /*if (*end == '_') goto bigparse;*/
    if (badcheck) {
      if (end == str) goto bad; /* no number */
      while (*end && ISSPACE(*end)) end++;
      if (*end) goto bad;        /* trailing garbage */
    }

    /*if (POSFIXABLE(val)) {*/
      if (sign) return mrb_fixnum_value(val);
      else {
        long result = -(long)val;
        return mrb_fixnum_value(result);
      }
    /*}
    else {
      mrb_value big = mrb_uint2big(val);
      RBIGNUM(big)->sign = sign;
      return bignorm(big);
    }*/
  //}
/*bigparse:
  len = (len/BITSPERDIG)+1;
  if (badcheck && *str == '_') goto bad;

  z = bignew(len, sign);
  zds = BDIGITS(z);
  for (i=len;i--;) zds[i]=0;
  while ((c = *str++) != 0) {
    if (c == '_') {
      if (nondigit) {
        if (badcheck) goto bad;
        break;
      }
      nondigit = c;
      continue;
    }
    else if ((c = conv_digit(c)) < 0) {
      break;
    }
    if (c >= base) break;
    nondigit = 0;
    i = 0;
    num = c;
    for (;;) {
      while (i<blen) {
        num += (BDIGIT_DBL)zds[i]*base;
        zds[i++] = BIGLO(num);
        num = BIGDN(num);
      }
      if (num) {
        blen++;
        continue;
      }
      break;
    }
  }
  if (badcheck) {
    str--;
    if (s+1 < str && str[-1] == '_') goto bad;
    while (*str && ISSPACE(*str)) str++;
    if (*str) {
*/
bad:
      //mrb_invalid_str(s, "Integer");
      printf("Integer");
  /*  }
  }
  return bignorm(z);*/
  return mrb_fixnum_value(0);
}
char *
mrb_string_value_cstr(mrb_state *mrb, mrb_value *ptr)
{
  struct RString *ps = mrb_str_ptr(*ptr);
  char *s = ps->buf;

  if (!s || ps->len != strlen(s)) {
    mrb_raise(mrb, mrb->eArgumentError_class, "string contains null byte");
  }
  return s;
}

mrb_value
mrb_str_to_inum(mrb_state *mrb, mrb_value str, int base, int badcheck)
{
  char *s;
  size_t len;
  struct RString *ps = mrb_str_ptr(str);

  //StringValue(str);
  mrb_string_value(mrb, &str);
  if (badcheck) {
    //s = StringValueCStr(str);
    s = mrb_string_value_cstr(mrb, &str);
  }
  else {
    s = ps->buf;
  }
  if (s) {
    len = ps->len;
    if (s[len]) {    /* no sentinel somehow */
      //char *p = ALLOCA_N(char, len+1);
      char *p = mrb_malloc(mrb, len+1);

      //MEMCPY(p, s, char, len);
      memcpy(p, s, sizeof(char)*len);
      p[len] = '\0';
      s = p;
    }
  }
  return mrb_cstr_to_inum(mrb, s, base, badcheck);
}

/* 15.2.10.5.38 */
/*
 *  call-seq:
 *     str.to_i(base=10)   => integer
 *
 *  Returns the result of interpreting leading characters in <i>str</i> as an
 *  integer base <i>base</i> (between 2 and 36). Extraneous characters past the
 *  end of a valid number are ignored. If there is not a valid number at the
 *  start of <i>str</i>, <code>0</code> is returned. This method never raises an
 *  exception.
 *
 *     "12345".to_i             #=> 12345
 *     "99 red balloons".to_i   #=> 99
 *     "0a".to_i                #=> 0
 *     "0a".to_i(16)            #=> 10
 *     "hello".to_i             #=> 0
 *     "1100101".to_i(2)        #=> 101
 *     "1100101".to_i(8)        #=> 294977
 *     "1100101".to_i(10)       #=> 1100101
 *     "1100101".to_i(16)       #=> 17826049
 */
static mrb_value
mrb_str_to_i(mrb_state *mrb, mrb_value self)
{
  mrb_value *argv;
  int argc;
  //mrb_value b;
  int base;

  //mrb_scan_args(argc, *argv, "01", &b);
  mrb_get_args(mrb, "*", &argv, &argc);
  if (argc == 0)
    base = 10;
  else
    base = mrb_fixnum(argv[0]);

  if (base < 0) {
    mrb_raise(mrb, mrb->eArgumentError_class, "illegal radix %d", base);
  }
  return mrb_str_to_inum(mrb, self, base, 0/*Qfalse*/);
}

double
mrb_cstr_to_dbl(mrb_state *mrb, const char * p, int badcheck)
{
  const char *q;
  char *end;
  double d;
//  const char *ellipsis = "";
//  int w;
#define OutOfRange() (((w = end - p) > 20) ? (w = 20, ellipsis = "...") : (ellipsis = ""))

  if (!p) return 0.0;
  q = p;
  if (badcheck) {
    while (ISSPACE(*p)) p++;
  }
  else {
    while (ISSPACE(*p) || *p == '_') p++;
  }
  //errno = 0;
  d = strtod(p, &end);
  /*if (errno == ERANGE) {
    OutOfRange();
    mrb_warn("Float %.*s%s out of range", w, p, ellipsis);
    errno = 0;
  }*/
  if (p == end) {
    if (badcheck) {
bad:
      //mrb_invalid_str(q, "Float()");
      printf("Float()\n");
    }
    return d;
  }
  if (*end) {
    //char *buf = ALLOCA_N(char, strlen(p)+1);
    char *buf = mrb_malloc(mrb, strlen(p)+1);
    char *n = buf;

    while (p < end) *n++ = *p++;
    while (*p) {
      if (*p == '_') {
        /* remove underscores between digits */
        if (badcheck) {
          if (n == buf || !ISDIGIT(n[-1])) goto bad;
          ++p;
          if (!ISDIGIT(*p)) goto bad;
        }
        else {
          while (*++p == '_');
          continue;
        }
      }
      *n++ = *p++;
    }
    *n = '\0';
    p = buf;
    d = strtod(p, &end);
    /*if (errno == ERANGE) {
      OutOfRange();
      mrb_warn("Float %.*s%s out of range", w, p, ellipsis);
      errno = 0;
    }*/
    if (badcheck) {
      if (!end || p == end) goto bad;
      while (*end && ISSPACE(*end)) end++;
      if (*end) goto bad;
    }
  }
  /*if (errno == ERANGE) {
    errno = 0;
    OutOfRange();
    mrb_raise(mrb, mrb->eArgumentError_class, "Float %.*s%s out of range", w, q, ellipsis);
  }*/
  return d;
}

double
mrb_str_to_dbl(mrb_state *mrb, mrb_value str, int badcheck)
{
  char *s;
  size_t len;
  struct RString *ps;

  //StringValue(str);
  mrb_string_value(mrb, &str);
  ps = mrb_str_ptr(str);
  s = ps->buf;
  len = ps->len;
  if (s) {
    if (s[len]) {    /* no sentinel somehow */
      //char *p = ALLOCA_N(char, len+1);
      char *p = mrb_malloc(mrb, len+1);

      //MEMCPY(p, s, char, len);
      memcpy(p, s, sizeof(char)*len);
      p[len] = '\0';
      s = p;
    }
    if (badcheck && len != strlen(s)) {
      mrb_raise(mrb, mrb->eArgumentError_class, "string for Float contains null byte");
    }
  }
  return mrb_cstr_to_dbl(mrb, s, badcheck);
}

/* 15.2.10.5.39 */
/*
 *  call-seq:
 *     str.to_f   => float
 *
 *  Returns the result of interpreting leading characters in <i>str</i> as a
 *  floating point number. Extraneous characters past the end of a valid number
 *  are ignored. If there is not a valid number at the start of <i>str</i>,
 *  <code>0.0</code> is returned. This method never raises an exception.
 *
 *     "123.45e1".to_f        #=> 1234.5
 *     "45.67 degrees".to_f   #=> 45.67
 *     "thx1138".to_f         #=> 0.0
 */
static mrb_value
mrb_str_to_f(mrb_state *mrb, mrb_value self)
{
  //return mrb_float_new(mrb_str_to_dbl(self, 0/*Qfalse*/));
  return mrb_float_value(mrb_str_to_dbl(mrb, self, 0/*Qfalse*/));
}

/* 15.2.10.5.40 */
/*
 *  call-seq:
 *     str.to_s     => str
 *     str.to_str   => str
 *
 *  Returns the receiver.
 */
static mrb_value
mrb_str_to_s(mrb_state *mrb, mrb_value self)
{
  //if (mrb_obj_class(self) != mrb_cString) {
  if (mrb_obj_class(mrb, self) != mrb->string_class) {
    return mrb_str_dup(mrb, self);
  }
  return self;
}

/* 15.2.10.5.43 */
/*
 *  call-seq:
 *     str.upcase!   => str or nil
 *
 *  Upcases the contents of <i>str</i>, returning <code>nil</code> if no changes
 *  were made.
 */
static mrb_value
mrb_str_upcase_bang(mrb_state *mrb, mrb_value self)
{
  char *s, *send;
  int modify = 0;
  struct RString *ps = mrb_str_ptr(self);

  mrb_str_modify(self);
  s = ps->buf;
  send = s + ps->len;
  while (s < send) {
    if (ismbchar(*s)) {
      s+=mbclen(*s) - 1;
    }
    else if (ISLOWER(*s)) {
      *s = toupper(*s);
      modify = 1;
    }
    s++;
  }

  if (modify) return self;
  return mrb_nil_value();
}

/* 15.2.10.5.42 */
/*
 *  call-seq:
 *     str.upcase   => new_str
 *
 *  Returns a copy of <i>str</i> with all lowercase letters replaced with their
 *  uppercase counterparts. The operation is locale insensitive---only
 *  characters ``a'' to ``z'' are affected.
 *
 *     "hEllO".upcase   #=> "HELLO"
 */
static mrb_value
mrb_str_upcase(mrb_state *mrb, mrb_value self)
{
  mrb_value str;

  str = mrb_str_dup(mrb, self);
  mrb_str_upcase_bang(mrb, str);
  return str;
}

#undef SIGN_EXTEND_CHAR
#if __STDC__
# define SIGN_EXTEND_CHAR(c) ((signed char)(c))
#else  /* not __STDC__ */
/* As in Harbison and Steele.  */
# define SIGN_EXTEND_CHAR(c) ((((unsigned char)(c)) ^ 128) - 128)
#endif
#define is_identchar(c) (SIGN_EXTEND_CHAR(c)!=-1&&(ISALNUM(c) || (c) == '_' || ismbchar(c)))

static int
is_special_global_name(m)
    const char *m;
{
    switch (*m) {
      case '~': case '*': case '$': case '?': case '!': case '@':
      case '/': case '\\': case ';': case ',': case '.': case '=':
      case ':': case '<': case '>': case '\"':
      case '&': case '`': case '\'': case '+':
      case '0':
        ++m;
        break;
      case '-':
        ++m;
        if (is_identchar(*m)) m += mbclen(*m);
        break;
      default:
        if (!ISDIGIT(*m)) return 0;
        do ++m; while (ISDIGIT(*m));
    }
    return !*m;
}

int
mrb_symname_p(const char *name)
{
    const char *m = name;
    int localid = FALSE;

    if (!m) return FALSE;
    switch (*m) {
      case '\0':
        return FALSE;

      case '$':
        if (is_special_global_name(++m)) return TRUE;
        goto id;

      case '@':
        if (*++m == '@') ++m;
        goto id;

      case '<':
        switch (*++m) {
          case '<': ++m; break;
          case '=': if (*++m == '>') ++m; break;
          default: break;
        }
        break;

      case '>':
        switch (*++m) {
          case '>': case '=': ++m; break;
        }
        break;

      case '=':
        switch (*++m) {
          case '~': ++m; break;
          case '=': if (*++m == '=') ++m; break;
          default: return FALSE;
        }
        break;

      case '*':
        if (*++m == '*') ++m;
        break;

      case '+': case '-':
        if (*++m == '@') ++m;
        break;

      case '|': case '^': case '&': case '/': case '%': case '~': case '`':
        ++m;
        break;

      case '[':
        if (*++m != ']') return FALSE;
        if (*++m == '=') ++m;
        break;

      default:
        localid = !ISUPPER(*m);
id:
        if (*m != '_' && !ISALPHA(*m) && !ismbchar(*m)) return FALSE;
        while (is_identchar(*m)) m += mbclen(*m);
        if (localid) {
            switch (*m) {
              case '!': case '?': case '=': ++m;
            }
        }
        break;
    }
    return *m ? FALSE : TRUE;
}

/*
 *  call-seq:
 *     str.dump   => new_str
 *
 *  Produces a version of <i>str</i> with all nonprinting characters replaced by
 *  <code>\nnn</code> notation and all special characters escaped.
 */

mrb_value
mrb_str_dump(mrb_state *mrb, mrb_value str)
{
    long len;
    char *p, *pend;
    char *q, *qend;
    mrb_value result;

    len = 2;                        /* "" */
    p = RSTRING(str)->buf; pend = p + RSTRING(str)->len;
    while (p < pend) {
        char c = *p++;
        switch (c) {
          case '"':  case '\\':
          case '\n': case '\r':
          case '\t': case '\f':
          case '\013': case '\010': case '\007': case '\033':
            len += 2;
            break;

          case '#':
            len += IS_EVSTR(p, pend) ? 2 : 1;
            break;

          default:
            if (ISPRINT(c)) {
                len++;
            }
            else {
                len += 4;                /* \nnn */
            }
            break;
        }
    }

    //result = mrb_str_new5(str, 0, len);
    result = mrb_str_new(mrb, 0, len);
    p = RSTRING(str)->buf; pend = p + RSTRING(str)->len;
    q = RSTRING(result)->buf; qend = q + len;

    *q++ = '"';
    while (p < pend) {
        char c = *p++;

        if (c == '"' || c == '\\') {
            *q++ = '\\';
            *q++ = c;
        }
        else if (c == '#') {
            if (IS_EVSTR(p, pend)) *q++ = '\\';
            *q++ = '#';
        }
        else if (ISPRINT(c)) {
            *q++ = c;
        }
        else if (c == '\n') {
            *q++ = '\\';
            *q++ = 'n';
        }
        else if (c == '\r') {
            *q++ = '\\';
            *q++ = 'r';
        }
        else if (c == '\t') {
            *q++ = '\\';
            *q++ = 't';
        }
        else if (c == '\f') {
            *q++ = '\\';
            *q++ = 'f';
        }
        else if (c == '\013') {
            *q++ = '\\';
            *q++ = 'v';
        }
        else if (c == '\010') {
            *q++ = '\\';
            *q++ = 'b';
        }
        else if (c == '\007') {
            *q++ = '\\';
            *q++ = 'a';
        }
        else if (c == '\033') {
            *q++ = '\\';
            *q++ = 'e';
        }
        else {
            *q++ = '\\';
            sprintf(q, "%03o", c&0xff);
            q += 3;
        }
    }
    *q++ = '"';

    //OBJ_INFECT(result, str);
    return result;
}

mrb_value
mrb_tainted_str_new(mrb_state *mrb, const char *ptr, long len)
{
    mrb_value str = mrb_str_new(mrb, ptr, len);

    //OBJ_TAINT(str);
    return str;
}

char *
mrb_string_value_ptr(mrb_state *mrb, mrb_value ptr)
{
    mrb_value str = mrb_string_value(mrb, &ptr);
    return RSTRING_PTR(str);
}

void
mrb_str_setter(mrb_state *mrb, mrb_value val, mrb_sym id, mrb_value *var)
{
  if (!mrb_nil_p(val) && (mrb_type(val) != MRB_TT_STRING)) {
    mrb_raise(mrb, mrb->eTypeError_class, "value of %s must be String", mrb_sym2name(mrb, id));
  }
  *var = val;
}

mrb_value
rb_str_buf_append(mrb_state *mrb, mrb_value str, mrb_value str2)
{
    str_buf_cat(mrb, str, RSTRING(str2)->buf, RSTRING(str2)->len);
    //OBJ_INFECT(str, str2);
    return str;
}

mrb_value
mrb_str_append(mrb_state *mrb, mrb_value str, mrb_value str2)
{
  //StringValue(str2);
  mrb_string_value(mrb, &str2);
  //mrb_str_modify(str);
  if (RSTRING(str2)->len > 0) {
    //if (FL_TEST(str, STR_ASSOC)) {
        long len = RSTRING(str)->len+RSTRING(str2)->len;
        //REALLOC_N(RSTRING(str)->buf, char, len+1);
        RSTRING(str)->buf = mrb_realloc(mrb, RSTRING(str)->buf, len+1);
        memcpy(RSTRING(str)->buf + RSTRING(str)->len,
             RSTRING(str2)->buf, RSTRING(str2)->len);
        RSTRING(str)->buf[len] = '\0'; /* sentinel */
        RSTRING(str)->len = len;
        RSTRING(str)->capa = len;
    /*}
    else {
        return mrb_str_buf_append(mrb, str, str2);
    }*/
  }
  //OBJ_INFECT(str, str2);
  return str;
}

mrb_value
rb_str_buf_cat(mrb_state *mrb, mrb_value str, const char * ptr, long len)
{
  if (len == 0) return str;
  if (len < 0) {
    mrb_raise(mrb, mrb->eArgumentError_class, "negative string size (or size too big)");
  }
  return str_buf_cat(mrb, str, ptr, len);
}

mrb_value
mrb_str_cat(mrb_state *mrb, mrb_value str, const char *ptr, long len)
{
  if (len < 0) {
    mrb_raise(mrb, mrb->eArgumentError_class, "negative string size (or size too big)");
  }
  if (0/*FL_TEST(str, STR_ASSOC)*/) {
    mrb_str_modify(str);
    //REALLOC_N(RSTRING(str)->buf, char, RSTRING(str)->len+len+1);
    mrb_realloc(mrb, RSTRING(str)->buf, RSTRING(str)->len+len+1);
    memcpy(RSTRING(str)->buf + RSTRING(str)->len, ptr, len);
    RSTRING(str)->len += len;
    RSTRING(str)->buf[RSTRING(str)->len] = '\0'; /* sentinel */
    return str;
  }

  return rb_str_buf_cat(mrb, str, ptr, len);
}

mrb_value
mrb_str_cat2(mrb_state *mrb, mrb_value str, const char *ptr)
{
    return str_buf_cat(mrb, str, ptr, strlen(ptr));
}

mrb_value
mrb_str_new_frozen(mrb_state *mrb, mrb_value orig)
{
  mrb_value str;
  str = mrb_str_literal(mrb, orig);
  return str;
}

const unsigned char OnigEncAsciiToLowerCaseTable[] = {
  '\000', '\001', '\002', '\003', '\004', '\005', '\006', '\007',
  '\010', '\011', '\012', '\013', '\014', '\015', '\016', '\017',
  '\020', '\021', '\022', '\023', '\024', '\025', '\026', '\027',
  '\030', '\031', '\032', '\033', '\034', '\035', '\036', '\037',
  '\040', '\041', '\042', '\043', '\044', '\045', '\046', '\047',
  '\050', '\051', '\052', '\053', '\054', '\055', '\056', '\057',
  '\060', '\061', '\062', '\063', '\064', '\065', '\066', '\067',
  '\070', '\071', '\072', '\073', '\074', '\075', '\076', '\077',
  '\100', '\141', '\142', '\143', '\144', '\145', '\146', '\147',
  '\150', '\151', '\152', '\153', '\154', '\155', '\156', '\157',
  '\160', '\161', '\162', '\163', '\164', '\165', '\166', '\167',
  '\170', '\171', '\172', '\133', '\134', '\135', '\136', '\137',
  '\140', '\141', '\142', '\143', '\144', '\145', '\146', '\147',
  '\150', '\151', '\152', '\153', '\154', '\155', '\156', '\157',
  '\160', '\161', '\162', '\163', '\164', '\165', '\166', '\167',
  '\170', '\171', '\172', '\173', '\174', '\175', '\176', '\177',
  '\200', '\201', '\202', '\203', '\204', '\205', '\206', '\207',
  '\210', '\211', '\212', '\213', '\214', '\215', '\216', '\217',
  '\220', '\221', '\222', '\223', '\224', '\225', '\226', '\227',
  '\230', '\231', '\232', '\233', '\234', '\235', '\236', '\237',
  '\240', '\241', '\242', '\243', '\244', '\245', '\246', '\247',
  '\250', '\251', '\252', '\253', '\254', '\255', '\256', '\257',
  '\260', '\261', '\262', '\263', '\264', '\265', '\266', '\267',
  '\270', '\271', '\272', '\273', '\274', '\275', '\276', '\277',
  '\300', '\301', '\302', '\303', '\304', '\305', '\306', '\307',
  '\310', '\311', '\312', '\313', '\314', '\315', '\316', '\317',
  '\320', '\321', '\322', '\323', '\324', '\325', '\326', '\327',
  '\330', '\331', '\332', '\333', '\334', '\335', '\336', '\337',
  '\340', '\341', '\342', '\343', '\344', '\345', '\346', '\347',
  '\350', '\351', '\352', '\353', '\354', '\355', '\356', '\357',
  '\360', '\361', '\362', '\363', '\364', '\365', '\366', '\367',
  '\370', '\371', '\372', '\373', '\374', '\375', '\376', '\377',
};

const unsigned char OnigEncAsciiToUpperCaseTable[256] = {
  '\000', '\001', '\002', '\003', '\004', '\005', '\006', '\007',
  '\010', '\011', '\012', '\013', '\014', '\015', '\016', '\017',
  '\020', '\021', '\022', '\023', '\024', '\025', '\026', '\027',
  '\030', '\031', '\032', '\033', '\034', '\035', '\036', '\037',
  '\040', '\041', '\042', '\043', '\044', '\045', '\046', '\047',
  '\050', '\051', '\052', '\053', '\054', '\055', '\056', '\057',
  '\060', '\061', '\062', '\063', '\064', '\065', '\066', '\067',
  '\070', '\071', '\072', '\073', '\074', '\075', '\076', '\077',
  '\100', '\101', '\102', '\103', '\104', '\105', '\106', '\107',
  '\110', '\111', '\112', '\113', '\114', '\115', '\116', '\117',
  '\120', '\121', '\122', '\123', '\124', '\125', '\126', '\127',
  '\130', '\131', '\132', '\133', '\134', '\135', '\136', '\137',
  '\140', '\101', '\102', '\103', '\104', '\105', '\106', '\107',
  '\110', '\111', '\112', '\113', '\114', '\115', '\116', '\117',
  '\120', '\121', '\122', '\123', '\124', '\125', '\126', '\127',
  '\130', '\131', '\132', '\173', '\174', '\175', '\176', '\177',
  '\200', '\201', '\202', '\203', '\204', '\205', '\206', '\207',
  '\210', '\211', '\212', '\213', '\214', '\215', '\216', '\217',
  '\220', '\221', '\222', '\223', '\224', '\225', '\226', '\227',
  '\230', '\231', '\232', '\233', '\234', '\235', '\236', '\237',
  '\240', '\241', '\242', '\243', '\244', '\245', '\246', '\247',
  '\250', '\251', '\252', '\253', '\254', '\255', '\256', '\257',
  '\260', '\261', '\262', '\263', '\264', '\265', '\266', '\267',
  '\270', '\271', '\272', '\273', '\274', '\275', '\276', '\277',
  '\300', '\301', '\302', '\303', '\304', '\305', '\306', '\307',
  '\310', '\311', '\312', '\313', '\314', '\315', '\316', '\317',
  '\320', '\321', '\322', '\323', '\324', '\325', '\326', '\327',
  '\330', '\331', '\332', '\333', '\334', '\335', '\336', '\337',
  '\340', '\341', '\342', '\343', '\344', '\345', '\346', '\347',
  '\350', '\351', '\352', '\353', '\354', '\355', '\356', '\357',
  '\360', '\361', '\362', '\363', '\364', '\365', '\366', '\367',
  '\370', '\371', '\372', '\373', '\374', '\375', '\376', '\377',
};
#define ONIGENC_ASCII_CODE_TO_LOWER_CASE(c) OnigEncAsciiToLowerCaseTable[c]
#define ONIGENC_ASCII_CODE_TO_UPPER_CASE(c) OnigEncAsciiToUpperCaseTable[c]

#define mrb_isascii(c) ((unsigned long)(c) < 128)
int
mrb_tolower(int c)
{
    return mrb_isascii(c) ? ONIGENC_ASCII_CODE_TO_LOWER_CASE(c) : c;
}

int
mrb_toupper(int c)
{
    return mrb_isascii(c) ? ONIGENC_ASCII_CODE_TO_UPPER_CASE(c) : c;
}

/* 15.2.10.5.46(x) */
/*
 * call-seq:
 *   str.inspect   => string
 *
 * Returns a printable version of _str_, with special characters
 * escaped.
 *
 *    str = "hello"
 *    str[3] = 8
 *    str.inspect       #=> "hel\010o"
 */

mrb_value
mrb_str_inspect(mrb_state *mrb, mrb_value str)
{
    char *p, *pend;
    mrb_value result = mrb_str_new_cstr(mrb, "\"");//mrb_str_buf_new2("\"");
    char s[5];

    p = RSTRING(str)->buf; pend = p + RSTRING(str)->len;
    while (p < pend) {
      char c = *p++;
      int len;
      if (ismbchar(c) && p - 1 + (len = mbclen(c)) <= pend) {
          rb_str_buf_cat(mrb, result, p - 1, len);
          p += len - 1;
      }
      else if (c == '"'|| c == '\\' || (c == '#' && IS_EVSTR(p, pend))) {
          s[0] = '\\'; s[1] = c;
          rb_str_buf_cat(mrb, result, s, 2);
      }
      else if (ISPRINT(c)) {
          s[0] = c;
          rb_str_buf_cat(mrb, result, s, 1);
      }
      else if (c == '\n') {
          s[0] = '\\'; s[1] = 'n';
          rb_str_buf_cat(mrb, result, s, 2);
      }
      else if (c == '\r') {
          s[0] = '\\'; s[1] = 'r';
          rb_str_buf_cat(mrb, result, s, 2);
      }
      else if (c == '\t') {
          s[0] = '\\'; s[1] = 't';
          rb_str_buf_cat(mrb, result, s, 2);
      }
      else if (c == '\f') {
          s[0] = '\\'; s[1] = 'f';
          rb_str_buf_cat(mrb, result, s, 2);
      }
      else if (c == '\013') {
          s[0] = '\\'; s[1] = 'v';
          rb_str_buf_cat(mrb, result, s, 2);
      }
      else if (c == '\010') {
          s[0] = '\\'; s[1] = 'b';
          rb_str_buf_cat(mrb, result, s, 2);
      }
      else if (c == '\007') {
          s[0] = '\\'; s[1] = 'a';
          rb_str_buf_cat(mrb, result, s, 2);
      }
      else if (c == 033) {
          s[0] = '\\'; s[1] = 'e';
          rb_str_buf_cat(mrb, result, s, 2);
      }
      else {
          sprintf(s, "\\%03o", c & 0377);
          mrb_str_buf_cat(mrb, result, s, strlen(s)); //mrb_str_buf_cat2(result, s);
      }
    }
    mrb_str_buf_cat(mrb, result, "\"", strlen("\"")); //mrb_str_buf_cat2(result, "\"");

    //OBJ_INFECT(result, str);
    return result;
}

/* ---------------------------*/
void
mrb_init_string(mrb_state *mrb)
{
  struct RClass *s;

  s = mrb->string_class = mrb_define_class(mrb, "String", mrb->object_class);
  MRB_SET_INSTANCE_TT(s, MRB_TT_STRING);
  mrb_include_module(mrb, s, mrb_class_get(mrb, "Comparable"));

  mrb_define_method(mrb, s, "+",               mrb_str_plus_m,          ARGS_REQ(1));              /* 15.2.10.5.2  */
  mrb_define_method(mrb, s, "bytesize",        mrb_str_bytesize,        ARGS_NONE());
  mrb_define_method(mrb, s, "size",            mrb_str_size,            ARGS_NONE());              /* 15.2.10.5.33 */
  mrb_define_method(mrb, s, "length",          mrb_str_size,            ARGS_NONE());              /* 15.2.10.5.26 */
  mrb_define_method(mrb, s, "*",               mrb_str_times,           ARGS_REQ(1));              /* 15.2.10.5.1  */
  mrb_define_method(mrb, s, "<=>",             mrb_str_cmp_m,           ARGS_REQ(1));              /* 15.2.10.5.3  */
  mrb_define_method(mrb, s, "==",              mrb_str_equal_m,         ARGS_REQ(1));              /* 15.2.10.5.4  */
  mrb_define_method(mrb, s, "=~",              mrb_str_match,           ARGS_REQ(1));              /* 15.2.10.5.5  */
  mrb_define_method(mrb, s, "[]",              mrb_str_aref_m,          ARGS_ANY());               /* 15.2.10.5.6  */
  mrb_define_method(mrb, s, "capitalize",      mrb_str_capitalize,      ARGS_NONE());              /* 15.2.10.5.7  */
  mrb_define_method(mrb, s, "capitalize!",     mrb_str_capitalize_bang, ARGS_REQ(1));              /* 15.2.10.5.8  */
  mrb_define_method(mrb, s, "chomp",           mrb_str_chomp,           ARGS_ANY());               /* 15.2.10.5.9  */
  mrb_define_method(mrb, s, "chomp!",          mrb_str_chomp_bang,      ARGS_ANY());               /* 15.2.10.5.10 */
  mrb_define_method(mrb, s, "chop",            mrb_str_chop,            ARGS_REQ(1));              /* 15.2.10.5.11 */
  mrb_define_method(mrb, s, "chop!",           mrb_str_chop_bang,       ARGS_REQ(1));              /* 15.2.10.5.12 */
  mrb_define_method(mrb, s, "downcase",        mrb_str_downcase,        ARGS_NONE());              /* 15.2.10.5.13 */
  mrb_define_method(mrb, s, "downcase!",       mrb_str_downcase_bang,   ARGS_NONE());              /* 15.2.10.5.14 */
  mrb_define_method(mrb, s, "each_line",       mrb_str_each_line,       ARGS_REQ(1));              /* 15.2.10.5.15 */
  mrb_define_method(mrb, s, "empty?",          mrb_str_empty,           ARGS_NONE());              /* 15.2.10.5.16 */
  mrb_define_method(mrb, s, "eql?",            mrb_str_eql,             ARGS_REQ(1));              /* 15.2.10.5.17 */
#ifdef INCLUDE_REGEXP
  mrb_define_method(mrb, s, "gsub",            mrb_str_gsub,            ARGS_REQ(1));              /* 15.2.10.5.18 */
  mrb_define_method(mrb, s, "gsub!",           mrb_str_gsub_bang,       ARGS_REQ(1));              /* 15.2.10.5.19 */
#endif
  mrb_define_method(mrb, s, "hash",            mrb_str_hash_m,          ARGS_REQ(1));              /* 15.2.10.5.20 */
  mrb_define_method(mrb, s, "include?",        mrb_str_include,         ARGS_REQ(1));              /* 15.2.10.5.21 */
  mrb_define_method(mrb, s, "index",           mrb_str_index_m,         ARGS_ANY());               /* 15.2.10.5.22 */
  mrb_define_method(mrb, s, "initialize",      mrb_str_init,            ARGS_REQ(1));              /* 15.2.10.5.23 */
  mrb_define_method(mrb, s, "initialize_copy", mrb_str_replace,         ARGS_REQ(1));              /* 15.2.10.5.24 */
  mrb_define_method(mrb, s, "intern",          mrb_str_intern,          ARGS_NONE());              /* 15.2.10.5.25 */
#ifdef INCLUDE_REGEXP
  mrb_define_method(mrb, s, "match",           mrb_str_match_m,         ARGS_REQ(1));              /* 15.2.10.5.27 */
#endif
  mrb_define_method(mrb, s, "replace",         mrb_str_replace,         ARGS_REQ(1));              /* 15.2.10.5.28 */
  mrb_define_method(mrb, s, "reverse",         mrb_str_reverse,         ARGS_NONE());              /* 15.2.10.5.29 */
  mrb_define_method(mrb, s, "reverse!",        mrb_str_reverse_bang,    ARGS_NONE());              /* 15.2.10.5.30 */
  mrb_define_method(mrb, s, "rindex",          mrb_str_rindex_m,        ARGS_ANY());               /* 15.2.10.5.31 */
#ifdef INCLUDE_REGEXP
  mrb_define_method(mrb, s, "scan",            mrb_str_scan,            ARGS_REQ(1));              /* 15.2.10.5.32 */
#endif
  mrb_define_method(mrb, s, "slice",           mrb_str_aref_m,          ARGS_ANY());               /* 15.2.10.5.34 */
  mrb_define_method(mrb, s, "split",           mrb_str_f_split,         ARGS_ANY());               /* 15.2.10.5.35 */
#ifdef INCLUDE_REGEXP
  mrb_define_method(mrb, s, "sub",             mrb_str_f_sub,           ARGS_REQ(1));              /* 15.2.10.5.36 */
  mrb_define_method(mrb, s, "sub!",            mrb_str_f_sub_bang,      ARGS_REQ(1));              /* 15.2.10.5.37 */
#endif
  mrb_define_method(mrb, s, "to_i",            mrb_str_to_i,            ARGS_ANY());               /* 15.2.10.5.38 */
  mrb_define_method(mrb, s, "to_f",            mrb_str_to_f,            ARGS_NONE());              /* 15.2.10.5.39 */
  mrb_define_method(mrb, s, "to_s",            mrb_str_to_s,            ARGS_NONE());              /* 15.2.10.5.40 */
  mrb_define_method(mrb, s, "to_sym",          mrb_str_intern,          ARGS_NONE());              /* 15.2.10.5.41 */
  mrb_define_method(mrb, s, "upcase",          mrb_str_upcase,          ARGS_REQ(1));              /* 15.2.10.5.42 */
  mrb_define_method(mrb, s, "upcase!",         mrb_str_upcase_bang,     ARGS_REQ(1));              /* 15.2.10.5.43 */

#ifdef INCLUDE_REGEXP
  mrb_define_method(mrb, s, "encoding",        mrb_obj_encoding,        ARGS_NONE());              /* 15.2.10.x.44 */
  mrb_define_method(mrb, s, "force_encoding",  mrb_str_force_encoding,  ARGS_REQ(1));              /* 15.2.10.x.45 */
#endif

  mrb_define_method(mrb, s, "inspect",         mrb_str_inspect,         ARGS_NONE());              /* 15.2.10.5.46(x) */
#if 1
  //  mrb_fs = mrb_nil_value(); // deleted by yamashita
  //  mrb_define_variable(mrb, "$;", &mrb_fs);// deleted by yamashita
  //  mrb_define_variable(mrb, "$-F", &mrb_fs);// deleted by yamashita
#endif
}


#endif //#ifndef INCLUDE_REGEXP
